#!/usr/bin/env python
"""Perform operations based on the mapping of a job to network topology
"""

import math
import warnings
import tokio.connectors.craysdb as craysdb
from . import jobinfo as jobinfo

def get_job_diameter(jobid, nodemap_cache_file=None, jobinfo_cache_file=None):
    """Calculate the diameter of a job

    An extremely crude way to reduce a job's node allocation into a scalar
    metric.  Assumes nodes are equally capable and fall on a 3D network;
    calculates the center of mass of the job's node positions.

    Args:
        jobid (str): A logical job id from which nodes are determined and their
            topological placement is determined
        nodemap_cache_file (str): Full path to the file containing the cached
            contents to be used to determine the node position map
        jobinfo_cache_file (str): Full path to the file containing the cached
            contents to be used to convert the job id into a node list

    Returns:
        dict: Contains three keys representing three ways in which a job's
        radius can be expressed.  Keys are:

            * job_min_radius: The smallest distance between the job's center of
              mass and a job node
            * job_max_radius: The largest distance between the job's center of
              mass and a job node
            * job_avg_radius: The average distance between the job's center of
              mass and all job nodes

    """
    node_list = jobinfo.get_job_nodes(jobid=jobid, cache_file=jobinfo_cache_file)
    proc_table = craysdb.CraySdbProc(cache_file=nodemap_cache_file)
    node_positions = []
    if len(node_list) == 0:
        warnings.warn("no valid job_info received from jobinfo.get_job_nodes()")
        return {}
    for jobnode in node_list:
        if not jobnode.startswith('nid'):
            warnings.warn("unable to parse jobnode '%s' for jobid '%s'" % (jobnode, jobid))
            return {}
        nid_num = int(jobnode.lstrip('nid'))
        node_x = proc_table[nid_num]['x_coord']
        node_y = proc_table[nid_num]['y_coord']
        node_z = proc_table[nid_num]['z_coord']
        node_positions.append((node_x, node_y, node_z))

    # Three dimensional topology
    center = [0.0, 0.0, 0.0]
    for node_position in node_positions:
        center[0] += node_position[0]
        center[1] += node_position[1]
        center[2] += node_position[2]

    center[0] /= float(len(node_positions))
    center[1] /= float(len(node_positions))
    center[2] /= float(len(node_positions))

    min_r = 10000.0
    max_r = 0.0
    avg_r = 0.0
    for node_position in node_positions:
        r2 = (node_position[0] - center[0])**2.0
        r2 += (node_position[1] - center[1])**2.0
        r2 += (node_position[2] - center[2])**2.0
        r = math.sqrt(r2)
        if r < min_r:
            min_r = r
        if r > max_r:
            max_r = r
        avg_r += r

    return {
        "job_min_radius": min_r,
        "job_max_radius": max_r,
        "job_avg_radius": avg_r / float(len(node_positions)),
    }



#!/usr/bin/env python
"""
Class and tools to generate TOKIO UMAMI plots
"""

import json
import datetime
import collections
import textwrap
import numpy
import pandas

DEFAULT_LINEWIDTH = 1
DEFAULT_LINECOLOR = "#853692"
DEFAULT_FONTSIZE = 12
DEFAULT_COLORSCALE = ['#DA0017', '#FD6A07', '#40A43A', '#2C69A9']
DEFAULT_FIGSIZE = (6.0, 12.0 / 9.0)

class Umami(collections.OrderedDict):
    """
    Subclass of dictionary that stores all of the data needed to generate an
    UMAMI diagram.  It is keyed by a metric name, and values are UmamiMetric
    objects which contain timestamps (x values) and measurements (y values)
    """
#   def __init__(self, *args, **kwargs):
#       super(Umami, self).__init__(*args, **kwargs)

    def to_dict(self):
        """
        Convert this object (and all of its constituent UmamiMetric objects)
        into a dictionary
        """
        return {k: v.__dict__ for k, v in self.items()}

    def _to_dict_for_pandas(self, stringify_key=False):
        """
        Convert this object into a DataFrame, indexed by timestamp, with each
        column as a metric.  The Umami attributes (labels, etc) are not
        expressed.
        """
        to_df = {}
        for metric, measurement in self.items():
            for index, timestamp in enumerate(measurement.timestamps):
                if stringify_key:
                    key = str(timestamp)
                else:
                    key = timestamp

                if key not in to_df:
                    to_df[key] = {}
                to_df[key].update({metric: measurement.values[index]})
        return to_df

    def to_json(self):
        """Serialize self into a JSON string

        Returns:
            str: JSON representation of numerical data being plotted
        """
        return json.dumps(self._to_dict_for_pandas(stringify_key=True), indent=4, sort_keys=True)

    def to_dataframe(self):
        """Return a representation of self as pandas.DataFrame

        Returns:
            pandas.DataFrame: numerical representation of the values being
            plotted
        """
        return pandas.DataFrame.from_dict(self._to_dict_for_pandas(), orient='index')

    def plot(self, output_file=None,
             highlight_index=-1,
             linewidth=DEFAULT_LINEWIDTH,
             linecolor=DEFAULT_LINECOLOR,
             colorscale=DEFAULT_COLORSCALE,
             fontsize=DEFAULT_FONTSIZE,
             figsize=DEFAULT_FIGSIZE):
        """Create a graphical representation of the UMAMI object

        Args:
            output_file (str or None): save umami diagram to file of given name
            highlight_index (int): index of measurement to highlight
            linewidth (int): linewidth for both timeseries and boxplot lines
            linecolor (str): color of line in timeseries panels
            colorscale (list of str): colors to use for data below the 25th,
                50th, 75th, and 100th percentiles
            fontsize (int): font size for UMAMI labels
            figsize (tuple of float): x, y dimensions of a single UMAMI row;
                multiplied by len(self.keys()) to determine full diagram height

        Returns:
            list: List of matplotlib.axis.Axis objects corresponding to each
            panel in the UMAMI diagram
        """
        # import here because of various things that can break matplotlib on import
        import matplotlib.pyplot

        rows_to_plot = list(self.keys())
        fig = matplotlib.pyplot.figure()
        fig.set_size_inches(figsize[0], len(rows_to_plot) * figsize[1])

        # Required to adjust the column widths of our figure (width_ratios)
        gridspec = matplotlib.gridspec.GridSpec(
            len(rows_to_plot),   # how many rows to draw
            2,                   # how many columns to draw
            width_ratios=[4, 1]) # ratio of column widths

        # Get the full range of x so we can force all rows to share the same
        # x range in the presence of trailing/leading NaNs
        x_min = None
        x_max = None
        for measurement in self.values():
            this_min = min(measurement.timestamps)
            this_max = max(measurement.timestamps)
            if x_min is None or this_min < x_min:
                x_min = this_min
            if x_max is None or this_max > x_max:
                x_max = this_max

        # Draw UMAMI rows
        last_ax_ts = None
        row_num = None
        for measurement in self.values():
            if row_num is None:
                row_num = 0
            else:
                row_num += 1

            x_val = measurement.timestamps
            y_val = measurement.values

            ### first plot the timeseries of the given variable
            ax_ts = fig.add_subplot(gridspec[2*row_num])
            ax_ts.plot(x_val,
                       y_val,
                       linestyle='-',
                       marker='x',
                       linewidth=linewidth,
                       color=linecolor)

            # textwrap.wrap inserts line breaks into each label
            ax_ts.set_ylabel('\n'.join(textwrap.wrap(text=measurement.label,
                                                     width=15,
                                                     break_on_hyphens=True)),
                             fontsize=fontsize,
                             rotation=0,
                             horizontalalignment='right',
                             verticalalignment='center')
            ax_ts.grid()
            ax_ts.set_xlim(x_min, x_max)

            # blank out the labels for all subplots except the bottom-most one
            if row_num != len(rows_to_plot) - 1:
                ax_ts.set_xticklabels([])
            else:
                last_ax_ts = ax_ts
                # resize and rotate the labels for the timeseries plot
                for tick in ax_ts.xaxis.get_major_ticks():
                    tick.label.set_fontsize(fontsize)
                    tick.label.set_rotation(45)

            # also adjust the font size for the y labels
            for tick in ax_ts.yaxis.get_major_ticks():
                tick.label.set_fontsize(fontsize)

            # then plot the boxplot summary of the given variable
            ax_box = fig.add_subplot(gridspec[2*row_num + 1])
            y_box_data = numpy.array(y_val)
            y_box_mask = [True] * len(y_box_data)
            y_box_mask[highlight_index] = False
            y_box_data = y_box_data[y_box_mask]
            y_box_data = y_box_data[~numpy.isnan(y_box_data)]
            ax_box.boxplot(y_box_data, # note: do not include last measurement in boxplot
                           widths=0.70,
                           boxprops={'linewidth':linewidth},
                           medianprops={'linewidth':linewidth},
                           whiskerprops={'linewidth':linewidth},
                           capprops={'linewidth':linewidth},
                           flierprops={'linewidth':linewidth},
                           whis=[5, 95])

            # scale the extents of the y ranges a little for clarity
            orig_ylim = ax_ts.get_ylim()
            new_ylim = list(map(lambda a, b: a*(1 + b), orig_ylim, (-0.1, 0.1)))
            ax_ts.set_ylim(new_ylim)

            yticks = ax_ts.get_yticks().tolist()

            # the following is a heuristic to determine how close the topmost
            # tick label is to the edge of the plot.  if it's too close, blank
            # it out so it doesn't overlap with the bottom-most tick label
            # of the row above it
            critical_fraction = abs(1.0 - (yticks[-1] - new_ylim[0]) / (new_ylim[-1] - new_ylim[0]))
            if row_num > 0 and critical_fraction < 0.01:
                # note that setting one of the yticks to a string resets the
                # formatting so that the tick labels appear as floats.  since
                # we (hopefully) would get integral ticks otherwise, force
                # them to ints.  This will mess things up if the yrange is
                # very narrow and must be expressed as floats.
                yticks = list(map(int, yticks))
                yticks[-1] = " "
                ax_ts.set_yticklabels(yticks)

            # lock in the y range to match the timeseries plot, just in case
            ax_box.set_ylim(ax_ts.get_ylim())

            # determine the color of our highlights based on quartile
            percentiles = [numpy.nanpercentile(y_val[0:-1], percentile)
                           for percentile in (25, 50, 75, 100)]

            color_index = 0
            for color_index, percentile in enumerate(percentiles):
                if y_val[highlight_index] <= percentile:
                    break

            if measurement.big_is_good:
                highlight_color = colorscale[color_index]
            else:
                highlight_color = colorscale[(1+color_index)*-1]

            # highlight the latest measurement on the timeseries plot
            x_last = matplotlib.dates.date2num(x_val[highlight_index])
            x_2nd_last = matplotlib.dates.date2num(x_val[highlight_index-1])
            ax_ts.plot([x_2nd_last, x_last],
                       [y_val[highlight_index-1], y_val[highlight_index]],
                       linestyle='-',
                       color=highlight_color,
                       linewidth=linewidth*2.0)
            ax_ts.plot([x_last], [y_val[highlight_index]],
                       marker='*',
                       color=highlight_color,
                       markersize=15)

            # where does this last data point lie on the distribution?
            ax_box.plot([0, 2],
                        [y_val[highlight_index], y_val[highlight_index]],
                        linestyle='--',
                        color=highlight_color,
                        linewidth=2.0,
                        zorder=10)

            # blank out all labels
            ax_box.set_yticklabels([""])
            ax_box.set_xticklabels([""])
            ax_box.yaxis.grid()

        fig.subplots_adjust(hspace=0.0, wspace=0.0)
        fig.autofmt_xdate()
        last_ax_ts.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%b %d'))

        if output_file is not None:
            fig.savefig(output_file, bbox_inches="tight")

        return fig.axes

class UmamiMetric(object):
    """A single row of an UMAMI diagram.
    
    Logically contains timeseries data from a single connector, where the
    `timestamps` attribute is a list of timestamps (seconds since epoch), and
    the 'values' attribute is a list of values corresponding to each timestamp.
    The number of timestamps and attributes must always be the same.
    """
    def __init__(self, timestamps, values, label, big_is_good=True):
        # If we are given pandas.Series, convert them to lists, then copy.
        # Otherwise, just copy the list-like inputs.
        if isinstance(timestamps, pandas.Series):
            self.timestamps = timestamps.tolist()[:]
        else:
            self.timestamps = timestamps[:]

        if isinstance(values, pandas.Series):
            self.values = values.tolist()[:]
        else:
            self.values = values[:]

        self.label = label
        self.big_is_good = big_is_good
        if len(self.timestamps) != len(self.values):
            raise Exception('timestamps and values must be of equal length')

    def to_json(self):
        """Create JSON-encoded string representation of self

        Returns:
            str: JSON-encoded representation of values stored in UmamiMetric
        """
        return json.dumps(self.__dict__, default=_serialize_datetime)

    def append(self, timestamp, value):
        """
        Can only add values along with a timestamp.
        """
        self.timestamps.append(timestamp)
        self.values.append(value)

    def pop(self):
        """
        Analogous to the list .pop() method.
        """
        timestamp = self.timestamps.pop()
        value = self.values.pop()
        return timestamp, value

def _serialize_datetime(obj):
    """
    Special serializer function that converts datetime into something that can
    be encoded in json
    """
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return (obj - datetime.datetime.utcfromtimestamp(0)).total_seconds()
    raise TypeError("Type %s not serializable" % type(obj))



"""
Summarize the contents of a TOKIO TimeSeries (TTS) HDF5 file generated by
:meth:`tokio.timeseries.TimeSeries.commit_dataset`.  This will eventually be
merged with the functionality provided by the summarize_h5lmt command-line tool.
"""

import json
import datetime
import argparse
import tokio.common
import tokio.timeseries
import tokio.connectors.hdf5

def humanize_units(byte_count, divisor=1024.0):
    """
    Convert a raw byte count into human-readable base2 units
    """
    units = ["bytes", "KiB", "MiB", "GiB", "TiB"]
    result = byte_count
    index = 0
    while index < len(units) - 1:
        new_result = result / divisor
        if new_result < 1.0:
            break
        else:
            index += 1
            result = new_result

    return result, units[index]

def summarize_tts_hdf5(hdf5_file):
    """
    Generate summary data based on the contents of TOKIO timeseries HDF5 file
    """
    read_bytes = hdf5_file['/datatargets/readbytes'][:, :].sum()
    write_bytes = hdf5_file['/datatargets/writebytes'][:, :].sum()

    # readrates and writerates come via the same collectd message, so if one is
    # missing, both are missing
    values = hdf5_file['/datatargets/readbytes'][:, :]
    num_missing = tokio.connectors.hdf5.missing_values(values).sum()
    total = values.shape[0] * values.shape[1]

    # find the row offset containing the first and last nonzero data
    first_time_idx = -1
    last_time_idx = -1
    nonzero_rows = tokio.connectors.hdf5.missing_values(values, inverse=True).sum(axis=1)
    for index, value in enumerate(nonzero_rows):
        if first_time_idx < 0 and value > 0:
            first_time_idx = index
        if value > 0:
            last_time_idx = index

    return {
        'read_bytes': read_bytes,
        'write_bytes': write_bytes,
        'missing_pts': num_missing,
        'total_pts': total,
        'missing_pct': (100.0 * float(num_missing) / total),
        'first_nonzero_idx': first_time_idx,
        'last_nonzero_idx': last_time_idx,
    }

def print_tts_hdf5_summary(results):
    """
    Format and print the summary data calculated by summarize_tts_hdf5()
    """
    print("Data Read:            %5.1f %s" % humanize_units(results['read_bytes']))
    print("Data Written:         %5.1f %s" % humanize_units(results['write_bytes']))
    print("Missing data points:  %9d" % results['missing_pts'])
    print("Expected data points: %9d" % results['total_pts'])
    print("Percent data missing: %8.1f%%" % results['missing_pct'])
    print("First non-empty row:  %9d" % results['first_nonzero_idx'])
    print("Last non-empty row:   %9d" % results['last_nonzero_idx'])

def summarize_timesteps(hdf5_file):
    """Summarizes total read/write bytes at each timestamp.

    Summarizes read/write bytes for each time step using the HDF5 interface
    instead of converting to a DataFrame or TimeSeries first.  Returns a dict
    of form::

        {
            "1546761600": {
                "read_bytes": 6135848142.0,
                "write_bytes": 6135848142.0
            },
            "1546761630": {
                "read_bytes": 5261439143.0,
                "write_bytes": 6135848142.0
            },
            "1546761660": {
                "read_bytes": 4321548241.0
                "write_bytes": 6135848142.0,
            },
            ...
        }

    """
    datasets = {
        '/datatargets/writebytes': 'write_bytes',
        '/datatargets/readbytes': 'read_bytes'
    }
    results = {}
    for dataset_name in datasets.keys():
        timestamps = hdf5_file.get_timestamps(dataset_name)[...]
        sum_bytes = hdf5_file[dataset_name][:, :].sum(axis=1)
        for index, timestamp in enumerate(timestamps):
            output_key = datasets.get(dataset_name)
            output_val = sum_bytes[index]
            timestamp_str = str(timestamp)
            if timestamp_str not in results:
                results[timestamp_str] = {}
            results[timestamp_str][output_key] = output_val

    return results

def print_timestep_summary(summary):
    """
    Format and print the summary data calculated by summarize_timesteps()
    """
    for timestamp, values in summary.items():
        print("%12s %14.2f read, %14.2f written" % (
            datetime.datetime.fromtimestamp(float(timestamp)),
            values.get('read_bytes', 0),
            values.get('write_bytes', 0)))

def summarize_columns(hdf5_file):
    """
    Summarize read/write bytes for each column
    """
    results = {}
    for index, column_name in enumerate(list(hdf5_file.get_columns('/datatargets/readbytes'))):
        if column_name not in results:
            results[column_name] = {}
        results[column_name]['read_bytes'] = hdf5_file['/datatargets/readbytes'][:, index].sum()

    for index, column_name in enumerate(list(hdf5_file.get_columns('/datatargets/writebytes'))):
        if column_name not in results:
            results[column_name] = {}
        results[column_name]['write_bytes'] = hdf5_file['/datatargets/writebytes'][:, index].sum()

    return results

def print_column_summary(results):
    """
    Format and print the summary data calculated by summarize_columns()
    """
    for column_name, values in results.items():
        print("%12s %14.2f read, %14.2f written" % (
            column_name,
            values.get('read_bytes', 0),
            values.get('write_bytes', 0)))

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=str, help="HDF5 file to summarize")
    parser.add_argument('-j', '--json', action='store_true', help='output as json')
    parser.add_argument('--timesteps', action='store_true', help='print a summary at each timestep')
    parser.add_argument('--columns', action='store_true', help='print a summary of each column')
    args = parser.parse_args(argv)

    hdf5_file = tokio.connectors.hdf5.Hdf5(args.file, 'r')
    results = {
        'total': summarize_tts_hdf5(hdf5_file),
    }

    if args.timesteps:
        results['timesteps'] = summarize_timesteps(hdf5_file)
    if args.columns:
        results['columns'] = summarize_columns(hdf5_file)

    if args.json:
        print(json.dumps(results, indent=4, sort_keys=True, cls=tokio.common.JSONEncoder))
    else:
        print_tts_hdf5_summary(results['total'])
        if 'timesteps' in results:
            print_timestep_summary(results['timesteps'])
        if 'columns' in results:
            print_column_summary(results['columns'])



"""
Dump a lot of data out of ElasticSearch using the Python API and native
scrolling support.

Instantiates a :class:`tokio.connectors.collectd_es.CollectdEs` object and
relies on the :meth:`tokio.connectors.collectd_es.CollectdEs.query_timeseries`
method to populate a data structure that is then serialized to JSON.
"""

import sys
import gzip
import json
import time
import datetime
import argparse
import warnings
import mimetypes

import tokio.debug
import tokio.connectors.collectd_es

DATE_FMT = "%Y-%m-%dT%H:%M:%S"

def main(argv=None):
    """Entry point for the CLI interface
    """
    warnings.simplefilter('always', UserWarning) # One warning per invalid file

    # Parse CLI options
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("query_start", type=str,
                        help="start time of query in %s format" % DATE_FMT)
    parser.add_argument("query_end", type=str,
                        help="end time of query in %s format" % DATE_FMT)
    parser.add_argument('--debug', action='store_true',
                        help="produce debug messages")
    parser.add_argument('--timeout', type=int, default=30,
                        help='ElasticSearch timeout time (default: 30)')
    parser.add_argument('--input', type=str, default=None,
                        help="use cached output from previous ES query")
    parser.add_argument("-o", "--output", type=str, default=None,
                        help="output file")
    parser.add_argument('-h', '--host', type=str, default="localhost",
                        help="hostname of ElasticSearch endpoint (default: localhost)")
    parser.add_argument('-p', '--port', type=int, default=9200,
                        help="port of ElasticSearch endpoint (default: 9200)")
    parser.add_argument('-i', '--index', type=str, default='cori-collectd-*',
                        help='ElasticSearch index to query (default:cori-collectd-*)')
    parser.add_argument("-c", "--csv", action="store_true", help="return output in CSV format")
    args = parser.parse_args(argv)

    if args.debug:
        tokio.DEBUG = True

    # Convert CLI options into datetime
    try:
        query_start = datetime.datetime.strptime(args.query_start, DATE_FMT)
        query_end = datetime.datetime.strptime(args.query_end, DATE_FMT)
    except ValueError:
        sys.stderr.write("Start and end times must be in format %s\n" % DATE_FMT)
        raise

    # Basic input bounds checking
    if query_start >= query_end:
        raise Exception('query_start >= query_end')

    # Read input from a cached json file (generated previously) or by querying
    # Elasticsearch directly
    must = []
    if args.input is None:
        ### Try to connect
        esdb = tokio.connectors.collectd_es.CollectdEs(
            host=args.host,
            port=args.port,
            index=args.index,
            timeout=args.timeout)

        pages = None
        for plugin_query in [tokio.connectors.collectd_es.QUERY_CPU_DATA,
                             tokio.connectors.collectd_es.QUERY_DISK_DATA,
                             tokio.connectors.collectd_es.QUERY_MEMORY_DATA]:
            esdb.query_timeseries(plugin_query,
                            query_start,
                            query_end)
            if pages is None:
                pages = esdb.scroll_pages
            else:
                pages += esdb.scroll_pages

        tokio.debug.debug_print("Loaded results from %s:%s" % (args.host, args.port))
    else:
        esdb = tokio.connectors.collectd_es.CollectdEs.from_cache(args.input)
        # the following query is arbitrary but is required to parse the cached output
        esdb.query_disk(query_start, query_end)
        tokio.debug.debug_print("Loaded results from %s" % args.input)

    # Write output
    cache_file = args.output
    if cache_file is not None:
        print("Caching to %s" % cache_file)

    if args.csv:
        if cache_file is None:
            print(esdb.to_dataframe().to_csv())
        else:
            esdb.to_dataframe().to_csv(cache_file)
    else:
        esdb.save_cache(cache_file)



#!/usr/bin/env python
"""Connect to NERSC's Intel Data Center Tool for SSDs outputs

Processes and aggregates the output of Intel Data Center Tool for SSDs outputs
in the format generated by NERSC's daily script.  The NERSC infrastructure runs
ISDCT in a verbose way on every burst buffer node, then collects all the
resulting output text files from a node into a directory bearing that node's nid
(e.g., ``nid01234/*.txt``).  There is also an optional timestamp file contained
in the toplevel directory.  Also processes .tar.gz versions of these collected
metrics.
"""

import os
import re
import tarfile
import time
import datetime
import warnings
import numpy
import pandas
from tokio.common import isstr, recast_string
from . import common

REX_SERIAL_NO = r'(Intel SSD|SMART Attributes|SMART and Health Information).*(CVF[^ ]+-\d+)'

class NerscIsdct(common.CacheableDict):
    """Dictionary subclass that self-populates with ISDCT output data
    """
    def __init__(self, input_file):
        """Load the output of a NERSC ISDCT dump.

        Args:
            input_file (str): Path to either a directory or a tar(/gzipped)
                directory containing the output of NERSC's ISDCT collection
                script.
        """
        super(NerscIsdct, self).__init__(input_file=input_file)

        # synthesize metrics independent of the input format
        self._synthesize_metrics()

    def load(self):
        """Wrapper around the filetype-specific loader.

        Infer the type of input being given, dispatch the correct loading
        function, and populate keys/values.
        """
        if not os.path.exists(self.input_file):
            raise Exception("Input file %s does not exist" % self.input_file)

        try:
            # try load_native first, because load_native understands that
            # self.input_file may actually be a directory
            self.load_native()
        except tarfile.ReadError:
            try:
                self.load_json()
            except ValueError as error:
                raise ValueError(str(error) + " (is this a valid json file?)")

    def load_native(self):
        """Load ISDCT output from a tar(.gz).

        Load a collection of ISDCT outputs as created by the NERSC ISDCT script.
        Assume that ISDCT output files each contain a single output from a
        single invocation of the isdct tool, and outputs are grouped into
        directories named according to their nid numbers (e.g.,
        nid00984/somefile.txt).
        """
        timestamp_str = None
        min_mtime = None
        parsed_counters_list = []
        for (member_name, mtime, member_handle) in common.walk_file_collection(self.input_file):
            # is this a magic timestamp file?
            if 'timestamp_' in member_name:
                timestamp_str = member_name.split('_')[-1]
                continue

            # independently, track the earliest mtime we can find.  In the
            # future, we may want to set an mtime for each individual file_obj,
            # but for now we treat the whole dump as a single timestamped entity
            if min_mtime is None:
                min_mtime = mtime
            else:
                min_mtime = min(min_mtime, mtime)

            # infer the node name from the member's path
            nodename = _decode_nersc_nid(member_name)

            # process the member's contents, then add it to self
            try:
                parsed_counters = parse_counters_fileobj(
                    member_handle,
                    nodename=nodename)
            except:
                warnings.warn("Parsing error in %s" % member_name)
                raise
            if parsed_counters is not None:
                parsed_counters_list.append(parsed_counters)

        # If no timestamp file was found, use the earliest mtime as a guess
        if timestamp_str is None:
            timestamp = datetime.datetime.fromtimestamp(min_mtime)
        # Otherwise, the timestamp file is the ground truth
        else:
            timestamp = datetime.datetime.strptime(timestamp_str, "%Y%m%d%H%M%S")

        # Set the timestamp for each device serial number
        timestamp = int(time.mktime(timestamp.timetuple()))
        for parsed_counters in parsed_counters_list:
            if len(parsed_counters) > 1:
                warnings.warn("Multiple serial numbers detected in a single file_obj: " \
                              + str(list(parsed_counters.keys())))
            for counters in parsed_counters.values():
                counters['timestamp'] = timestamp

        merged_counters = _merge_parsed_counters(parsed_counters_list)
        for key, val in merged_counters.items():
            self.__setitem__(key, val)

    def to_dataframe(self, only_numeric=False):
        """Express self as a dataframe.

        Args:
            only_numeric (bool): Only output columns containing numeric data of
                True; otherwise, output all columns.

        Returns:
            pandas.DataFrame: Dataframe indexed by serial number and with ISDCT
            counters as columns
        """
        dataframe = pandas.DataFrame.from_dict(self, orient='index')
        dataframe.index.name = "serial_number"
        if only_numeric:
            numeric_cols = []
            for column in sorted(dataframe.keys()):
                ### don't print counters that are non-numeric
                if dataframe[column].dtype != numpy.int64 \
                and dataframe[column].dtype != numpy.float64:
                    continue

                ### don't print counters that are the same for all devices. dataframe.count > 1
                ### because we can't tell how many rows are unique if there's only one
                ### row
                if dataframe.shape[0] > 1 and len(dataframe[column].unique()) == 1:
                    continue

                numeric_cols.append(column)

            if 'node_name' in dataframe:
                numeric_cols = ['node_name'] + numeric_cols
            return dataframe[numeric_cols]
        else:
            return dataframe

    def diff(self, old_isdct, report_zeros=True):
        """Highlight differences between self and another NerscIsdct.

        Subtract each counter for each serial number in this object from its
        counterpart in ``old_isdct``.  Return the changes in each numeric counter
        and any serial numbers that have appeared or disappeared.

        Args:
            old_isdct (NerscIsdct): object with which we should be compared
            report_zeros (bool): If True, report all counters even if they
                showed no change.  Default is True.

        Returns:
            dict: Dictionary containing the following keys:
                * `added_devices` - device serial numbers which exist in self
                  but not old_isdct
                * `removed_devices` - device serial numbers which do not exist
                  in self but do in old_isdct
                * `devices` - dict keyed by device serial numbers and whose
                  values are dicts of keys whose values are the difference
                  between old_isdct and self
        """
        result = {
            'added_devices': [],
            'removed_devices': [],
            'devices': {},
        }
        existing_devices = set([])
        for serial_no, counters in self.items():
            existing_devices.add(serial_no)
            # new devices that are appearing for the first time
            if serial_no not in old_isdct:
                result['added_devices'].append(serial_no)
                continue

            # calculate the diff in each counter for this device
            diff_dict = {}
            for counter, value in counters.items():
                if counter not in old_isdct[serial_no]:
                    warnings.warn("Counter %s does not exist in old_isdct" % counter)

                # just highlight different strings, but ignore
                # endurance_analyzer (which can be numeric or string-like)
                elif isstr(value) and counter != "endurance_analyzer":
                    if old_isdct[serial_no][counter] != value:
                        diff_value = "+++%s ---%s" % (value, old_isdct[serial_no][counter])
                    else:
                        diff_value = ""
                    if report_zeros or diff_value != "":
                        diff_dict[counter] = diff_value

                # subtract numeric counters
                else:
                    try:
                        diff_value = value - old_isdct[serial_no][counter]
                    except TypeError:
                        ### endurance_analyzer can be either numeric (common
                        ### case) or an error string (if the drive is brand
                        ### new); just drop the counter in the non-numeric
                        ### case
                        if counter == 'endurance_analyzer':
                            continue
                        error = "incompatible numeric types for %s/%s: [%s] vs [%s]" % (
                            serial_no,
                            counter,
                            old_isdct[serial_no][counter],
                            value)
                        raise TypeError(error)
                    if report_zeros or diff_value != 0:
                        diff_dict[counter] = diff_value

            result['devices'][serial_no] = diff_dict

        # Look for serial numbers that used to exist but do not appear in self
        for serial_no in old_isdct:
            if serial_no not in existing_devices:
                result['removed_devices'].append(serial_no)

        return result

    def _synthesize_metrics(self):
        """Calculate additional metrics not presented by ISDCT.

        Calculates additional convenient metrics that are not directly presented
        by ISDCT, then adds the resulting key-value pairs to self.
        """

        for serial_no, counters in self.items():
            # native units are "kiloblocks," whatever that signifies
            for key in 'data_units_read', 'data_units_written':
                if key in counters:
                    self[serial_no][key + "_bytes"] = counters[key] * 1000 * 512

            # native units are in 32 MiB
            for key in 'smart_nand_bytes_written_raw', 'smart_host_bytes_written_raw':
                if key in counters:
                    self[serial_no][key.replace('_raw', '_bytes')] = \
                        counters[key] * 32 * 1024 * 1024

            # calculate WAF
            nand_writes = counters.get('smart_nand_bytes_written_raw')
            host_writes = counters.get('smart_host_bytes_written_raw')
            if nand_writes is not None and host_writes is not None and host_writes > 0:
                self[serial_no]['write_amplification_factor'] = \
                    float(nand_writes) / float(host_writes)

def _merge_parsed_counters(parsed_counters_list):
    """Merge ISDCT outputs into a single object.

    Aggregates counters from each record based on the NVMe device serial number,
    with redundant counters being overwritten.

    Args:
        parsed_counters_list (list): List of parsed ISDCT outputs as dicts.
            Each list element is a dict with a single key (a device serial
            number) and one or more values; each value is itself a dict of
            key-value pairs corresponding to ISDCT/SMART counters from that
            device.

    Returns:
        dict: Dict with keys given by all device serial numbers found in
            `parsed_counters_list` and whose values are a dict containing keys
            and values representing all unique keys across all elements of
            `parsed_counters_list`.
    """
    all_data = {}
    for parsed_counters in parsed_counters_list:
        if not parsed_counters:
            continue
        elif len(parsed_counters) > 1:
            raise Exception("Received multiple serial numbers from parse_dct_counters_file")
        else:
            device_sn = next(iter(parsed_counters))

        ### merge file's counter dict with any previous counters we've parsed
        if device_sn not in all_data:
            all_data[device_sn] = parsed_counters[device_sn]
        else:
            all_data[device_sn].update(parsed_counters[device_sn])

    ### attempt to figure out the type of each counter
    new_counters = []
    for device_sn, counters in all_data.items():
        for counter, value in counters.items():
            ### first, handle counters that do not have an obvious way to cast
            if counter in ("temperature", "throttle_status", "endurance_analyzer"):
                tokens = value.split(None, 1)
                if len(tokens) == 2:
                    new_value, unit = tokens
                else:
                    new_value, unit = tokens[0], None
            else:
                new_value = value
                unit = None

            new_value = recast_string(value)

            ### endurance_analyzer can be numeric or an error string
            if counter == 'endurance_analyzer':
                if isstr(new_value):
                    new_value = None
                    unit = None
                elif unit is None:
                    # Intel reports this counter multiple times using different print formats
                    unit = 'years'

            ### Insert the key-value pair and its unit of measurement (if available)
            all_data[device_sn][counter] = new_value
            if unit is not None:
                new_counters.append((device_sn, '%s_unit' % counter, unit))

    for (device_sn, counter, value) in new_counters:
        all_data[device_sn][counter] = value

    return all_data

def _decode_nersc_nid(path):
    """Convert path to ISDCT output into a nid.

    Given a path to some ISDCT output file, somehow figure out what the nid
    name for that node is.  This encoding is specific to the way NERSC collects
    and preserves ISDCT outputs.

    Args:
        path (str): path to an ISDCT output text file

    Returns:
        str: Node identifier (e.g., nid01234)
    """
    abs_path = os.path.abspath(path)
    nid_name = os.path.dirname(abs_path).split(os.sep)[-1]
    return nid_name

def _rekey_smart_buffer(smart_buffer):
    """Convert SMART values associated with one register into unique counters.

    Take a buffer containing smart values associated with one register and
    create unique counters.  Only necessary for older versions of ISDCT
    which did not output SMART registers in a standard "Key: value" text
    format.

    Args:
        smart_buffer (dict): SMART buffer as defined by parse_counters_fileobj()

    Returns:
        dict: unique key:value pairs whose key now includes distinguishing
            device-specific characteristics to avoid collision from other
            devices that generated SMART data
    """
    data = {}
    prefix = smart_buffer.get("description")
    if prefix is None:
        prefix = smart_buffer.get("_id")

    for key, val in smart_buffer.items():
        key = ("SMART_%s_%s" % (prefix, key.strip())).replace("-", "").replace(" ", "_")
        key = _normalize_key(key)
        data[key] = val.strip()
    return data

def _normalize_key(key):
    """Coerce all keys into a similar naming convention.

    Converts Intel's mix of camel-case and snake-case counters into all
    snake-case.  Contains some nasty acronym hacks that may require modification
    if/when Intel adds new funny acronyms that contain a mix of upper and lower
    case letters (e.g., SMBus and NVMe).

    Args:
        key (str): a string to normalize

    Returns:
        str: snake_case version of key
    """
    key = key.strip().replace(' ', '_').replace('-', '')
    key = key.replace("NVMe", "nvme").replace("SMBus", "smbus")
    last_letter = None
    new_key = ""
    letter_list = list(key)
    for index, letter in enumerate(letter_list):

        if index+1 < len(letter_list):
            next_letter = letter_list[index+1]
        else:
            next_letter = None

        # don't allow repeated underscores
        if letter == "_" and last_letter == "_":
            continue

        if last_letter is not None and last_letter != "_":
            # start of a camelcase word?
            if letter.isupper():
                # don't stick underscores in acronyms
                if not last_letter.isupper() and last_letter != "_":
                    new_key += "_"
                # unless this is the end of an acronym
                elif next_letter is not None and not next_letter.isupper() and next_letter != "_":
                    new_key += "_"
        new_key += letter
        last_letter = letter

    return new_key.lower()

def parse_counters_fileobj(fileobj, nodename=None):
    """Convert any output of ISDCT into key-value pairs.

    Reads the output of a file-like object which contains the output of a
    single isdct command.  Understands the output of the following options:

      * ``isdct show -smart`` (SMART attributes)
      * ``isdct show -sensor`` (device health sensors)
      * ``isdct show -performance`` (device performance metrics)
      * ``isdct show -a`` (drive info)

    Args:
        fileobj (file): file-like object containing the output of an ISDCT
            command
        nodename (str): name of node corresponding to `fileobj`, if known

    Returns:
        dict: dict of dicts keyed by the device serial number.
    """

    data = {}
    device_sn = None
    parse_mode = 0      # =0 for regular counters, 1 for SMART data
    smart_buffer = {}
    for line in fileobj.readlines():
        if not isstr(line):
            line = line.decode().strip()
        else:
            line = line.strip()

        if device_sn is None:
            rex_match = re.search(REX_SERIAL_NO, line)
            if rex_match is not None:
                device_sn = rex_match.group(2)
                if nodename is not None:
                    data['node_name'] = nodename
                if rex_match.group(1) == "Intel SSD" \
                or rex_match.group(1) == "SMART and Health Information":
                    parse_mode = 0
                elif rex_match.group(1) == "SMART Attributes":
                    parse_mode = 1
                else:
                    raise Exception("Unknown counter file format")
        elif parse_mode == 0 and ':' in line:
            key, val = line.split(':', 1)
            key = _normalize_key(key)
            data[key] = val.strip()
        elif parse_mode > 0 and ':' in line:
            key, val = line.split(':', 1)
            key = _normalize_key(key)
            smart_buffer[key] = val.strip()
        elif parse_mode > 0 and line.startswith('-') and line.endswith('-'):
            for key, val in _rekey_smart_buffer(smart_buffer).items():
                key = _normalize_key(key)
                data[key] = val
            smart_buffer = {'_id' : line.split()[1]}
    if parse_mode > 0: # flush the last SMART register
        for key, val in _rekey_smart_buffer(smart_buffer).items():
            key = _normalize_key(key)
            data[key] = val

    if device_sn is None:
        warnings.warn("Could not find device serial number in %s" % fileobj.name)
        return {}

    return {device_sn : data}



"""
Provide a CLI interface for :meth:`tokio.connectors.nersc_isdct.NerscIsdct.to_dataframe`
and :meth:`tokio.connectors.nersc_isdct.NerscIsdct.save_cache` methods.
"""

import argparse
import tokio.connectors.nersc_isdct

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("isdctfile", type=str, help="NERSC ISDCT dump to process")
    parser.add_argument("-j", "--json", action="store_true", default=True,
                        help="return output in JSON format")
    parser.add_argument("-c", "--csv", action="store_true", help="return output in CSV format")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    args = parser.parse_args(argv)

    # Read from a cache file
    isdct_data = tokio.connectors.nersc_isdct.NerscIsdct(args.isdctfile)

    # Serialize the object
    cache_file = args.output
    if cache_file is not None:
        print("Caching to %s" % cache_file)

    if args.csv:
        if cache_file is None:
            print(isdct_data.to_dataframe().to_csv())
        else:
            isdct_data.to_dataframe().to_csv(cache_file)
    elif args.json:
        isdct_data.save_cache(cache_file)
    else:
        raise Exception("No output format specified")



"""Retrieves ESnet SNMP counters and store them in TOKIO Timeseries format
"""

import sys
import datetime
import argparse
import warnings
import tokio.debug
import tokio.config
import tokio.timeseries
import tokio.connectors.esnet_snmp
import tokio.connectors.hdf5

DATE_FMT = "%Y-%m-%dT%H:%M:%S"
DATE_FMT_PRINT = "YYYY-MM-DDTHH:MM:SS"

SCHEMA_VERSION = "1"

def endpoint_name(endpoint, interface):
    """Create a single key from an endpoint, interface pair

    Args:
        endpoint (str): The name of an endpoint
        interface (str): The interface on the given endpoint

    Returns:
        str: A single key combining endpoint and interface
    """
    return "%s:%s" % (endpoint, interface)

class Archiver(dict):
    """A dictionary containing TimeSeries objects

    Contains the TimeSeries objects being populated from a remote data source.
    Implemented as a class so that a single object can store all of the
    TimeSeries objects that are generated by multiple method calls.
    """
    def __init__(self, query_start, query_end, interfaces, timestep, *args, **kwargs):
        """Initializes the archiver and stores its settings

        Args:
            query_start (datetime.datetime): Lower bound of time to be archived,
                inclusive
            query_end (datetime.datetime): Upper bound of time to be archived,
                inclusive
            interfaces (list of tuples): List of endpoints and interfaces to
                archive.  Each tuple is of the form (endpoint, interface).
            timestep (int): Number of seconds between successive data points.
                The ESnet service may not honor this request.
        """
        super(Archiver, self).__init__(*args, **kwargs)
        self.query_start = query_start
        self.query_end = query_end
        self.timestep = timestep
        self.interfaces = interfaces

        self.config = {
            'datatargets/readrates': {
                "units": "bytes/sec",
                "direction": "in",
            },
            'datatargets/writerates': {
                "units": "bytes/sec",
                "direction": "out",
            },
        }

        self.columns = [endpoint_name(tup[0], tup[1]) for tup in self.interfaces]

        self.schema = tokio.connectors.hdf5.SCHEMA.get(SCHEMA_VERSION)
        if self.schema is None:
            raise KeyError("Schema version %d is not known by connectors.hdf5" % SCHEMA_VERSION)

    def init_datasets(self, dataset_names):
        """Populate empty datasets within self

        Creates and attachs TimeSeries objects to self based on a given column list

        Args:
            dataset_names (list of str): keys corresponding to self.config
                defining which datasets are being initialized
        """
        for dataset_name in dataset_names:
            hdf5_dataset_name = self.schema.get(dataset_name)
            if hdf5_dataset_name is None:
                warnings.warn("Skipping %s (not in schema)" % dataset_name)
            else:
                self[dataset_name] = tokio.timeseries.TimeSeries(dataset_name=hdf5_dataset_name,
                                                                 start=self.query_start,
                                                                 end=self.query_end,
                                                                 timestep=self.timestep,
                                                                 num_columns=len(self.columns),
                                                                 column_names=self.columns)

    def finalize(self):
        """Convert datasets to deltas where necessary and tack on metadata

        Perform a few finishing actions to all datasets contained in self after
        they have been populated.  Such actions are configured entirely in
        self.config and require no external input.
        """
        self.set_timeseries_metadata(list(self.config.keys()))

    def set_timeseries_metadata(self, dataset_names):
        """Set metadata constants (version, units, etc) on datasets and groups

        Args:
            dataset_names (list of str): keys corresponding to self.config for
                the datasets whose metadata should be set
        """
        for dataset_name in dataset_names:
            if dataset_name in self:
                self[dataset_name].dataset_metadata.update({
                    'version': SCHEMA_VERSION,
                    'units': self.config[dataset_name]['units']
                })
                self[dataset_name].group_metadata.update({'source': 'esnet_snmp'})

    def archive(self, input_file=None):
        """Extract and encode data from ESnet's SNMP service

        Queries the ESnet SNMP REST service, interprets resulting data, and
        populates a dictionary of TimeSeries objects with those values.

        Args:
            esnetsnmp (tokio.connectors.esnet_snmp.EsnetSnmp): Connector instance
        """

        esnetsnmp = tokio.connectors.esnet_snmp.EsnetSnmp(
            start=self.query_start,
            end=self.query_end,
            input_file=input_file)

        self.init_datasets(self.config.keys())

        # Retrieve all counters from REST API if not loaded from cache
        if input_file is None:
            last_payload_ck = None
            for endpoint, interface in self.interfaces:
                for direction in 'in', 'out':
                    raw_result = esnetsnmp.get_interface_counters(
                        endpoint=endpoint,
                        interface=interface,
                        direction=direction,
                        agg_func='average',
                        interval=self.timestep)

                    # verify that the payload returned data and didn't hit an API rate limit
                    payload = raw_result.get('data')
                    if not payload:
                        raise RuntimeError("%s:%s:%s returned no data" % (endpoint, interface, direction))

                    # verify that the payload returned a consistent number of data points
                    this_payload_ck = len(payload)
                    if last_payload_ck is not None and this_payload_ck != last_payload_ck:
                        raise RuntimeError(
                            "%s:%s:%s returned %d entries, but previous contained %d"
                            % (endpoint, interface, direction, this_payload_ck, last_payload_ck))

        # Loop over all retrieved data
        for dataset_name, config in self.config.items():
            direction = config['direction']
            for endpoint, interface in self.interfaces:
                for timestamp_int, bytes_sec in esnetsnmp[endpoint][interface][direction].items():
                    timestamp = datetime.datetime.fromtimestamp(timestamp_int)
                    target_name = endpoint_name(endpoint, interface)
                    self[dataset_name].insert_element(
                        timestamp,
                        target_name,
                        bytes_sec)

def init_hdf5_file(datasets, init_start, init_end, hdf5_file):
    """
    Initialize the datasets at full dimensions in the HDF5 file if necessary
    """
    schema = tokio.connectors.hdf5.SCHEMA.get(SCHEMA_VERSION)
    for dataset_name, dataset in datasets.items():
        hdf5_dataset_name = schema.get(dataset_name)
        if hdf5_dataset_name is None:
            if '/_' not in dataset_name:
                warnings.warn("Dataset key %s is not in schema" % dataset_name)
            continue
        if hdf5_dataset_name not in hdf5_file:
            # attempt to convert dataset into a timeseries
            timeseries = hdf5_file.to_timeseries(dataset_name=hdf5_dataset_name)

            # if dataset -> timeseries failed, create and commit a new, empty timeseries
            if timeseries is None:
                timeseries = tokio.timeseries.TimeSeries(dataset_name=hdf5_dataset_name,
                                                         start=init_start,
                                                         end=init_end,
                                                         timestep=dataset.timestep,
                                                         num_columns=dataset.dataset.shape[1])
                hdf5_file.commit_timeseries(timeseries=timeseries)
            print("Initialized %s in %s with size %s" % (
                hdf5_dataset_name,
                hdf5_file.name,
                timeseries.dataset.shape))

def archive_esnet_snmp(init_start, init_end, interfaces, timestep, output_file, query_start, query_end, input_file=None):
    """Retrieves remote data and stores it in TOKIO time series format

    Given a start and end time, retrieves all of the relevant contents of a
    remote data source and encodes them in the TOKIO time series HDF5 data
    format.

    Args:
        init_start (datetime.datetime): The first timestamp to be included in
            the HDF5 file
        init_end (datetime.datetime): The timestamp following the last timestamp
            to be included in the HDF5 file.
        interfaces (list of tuples): List of (endpoint, interface) elements to
            query.
        timestep (int): Number of seconds between successive entries in the HDF5
            file to be created.
        output_file (str): Path to the file to be created.
        query_start (datetime.datetime): Time after which remote data should be
            retrieved, inclusive.
        query_end (datetime.datetime): Time before which remote data should be
            retrieved, inclusive.
        input_file (str or None): Path to a cached input.  If specified, the
            remote REST API will not be contacted and the contents of this file
            will be instead loaded.
    """
    datasets = Archiver(query_start=query_start, query_end=query_end, interfaces=interfaces, timestep=timestep)

    datasets.archive(input_file=input_file)

    datasets.finalize()

    with tokio.connectors.hdf5.Hdf5(output_file) as hdf5_file:
        hdf5_file.attrs['version'] = SCHEMA_VERSION

        init_hdf5_file(datasets, init_start, init_end, hdf5_file)

        for dataset in datasets.values():
            print("Writing out %s" % dataset.dataset_name)
            hdf5_file.commit_timeseries(dataset)

    tokio.debug.debug_print("Wrote output to %s" % output_file)

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, default=None,
                        help="input cache db file")
    parser.add_argument("-o", "--output", type=str, default='output.hdf5',
                        help="output file (default: output.hdf5)")
    parser.add_argument('--init-start', type=str, default=None,
                        help='first timestamp (inclusive) when creating new output file,' +
                        ' in %s format (default: same as start)' % DATE_FMT_PRINT)
    parser.add_argument('--init-end', type=str, default=None,
                        help='final timestamp (exclusive) when creating new output file,' +
                        ' in %s format (default: same as end)' % DATE_FMT_PRINT)
    parser.add_argument('--debug', action='store_true', help="produce debug messages")
    parser.add_argument('--timestep', type=int, default=30,
                        help='collection frequency, in seconds (default: 30)')
    parser.add_argument("query_start", type=str,
                        help="start time in %s format" % DATE_FMT_PRINT)
    parser.add_argument("query_end", type=str,
                        help="end time in %s format" % DATE_FMT_PRINT)
    parser.add_argument("endpoints", type=str,
                        help='endpoint(s) to process, either a center name'
                        + ' ("nersc") or comma-separated list of format'
                        + ' "endpoint0:if0,endpoint1:if1,..." etc')
    args = parser.parse_args(argv)

    if args.debug:
        tokio.debug.DEBUG = True

    # Convert CLI options into datetime
    try:
        query_start = datetime.datetime.strptime(args.query_start, DATE_FMT)
        query_end = datetime.datetime.strptime(args.query_end, DATE_FMT)
        init_start = query_start
        init_end = query_end
        if args.init_start:
            init_start = datetime.datetime.strptime(args.init_start, DATE_FMT)
        if args.init_end:
            init_end = datetime.datetime.strptime(args.init_end, DATE_FMT)
    except ValueError:
        sys.stderr.write("Start and end times must be in format %s\n" % DATE_FMT_PRINT)
        raise

    # Basic input bounds checking
    if query_start >= query_end:
        raise ValueError('query_start >= query_end')
    elif init_start >= init_end:
        raise ValueError('query_start >= query_end')
    elif args.timestep < 1:
        raise ValueError('--timestep must be > 0')

    # Build list of desired interfaces
    if ':' not in args.endpoints:
        query_args = tokio.config.CONFIG.get('esnet_snmp_interfaces', {}).get(args.endpoints, {}).copy()
    else:
        query_args = {}
        for kvpair in args.endpoints.split(","):
            key, value = kvpair.split(":", 1)
            if key not in query_args:
                query_args[key] = []
            query_args[key].append(value)
    interfaces = []
    for endpoint, ifs in query_args.items():
        interfaces += [(endpoint, interface) for interface in ifs]

    archive_esnet_snmp(
        init_start=init_start,
        init_end=init_end,
        timestep=args.timestep,
        interfaces=interfaces,
        output_file=args.output,
        query_start=query_start,
        query_end=query_end,
        input_file=args.input)



#!/usr/bin/env python
"""
Connectors for the Lustre `lfs df` and `lctl dl -t` commands to determine the
health of Lustre file systems from the clients' perspective.
"""

import re
from tokio.connectors.common import SubprocessOutputDict

# Only try to match osc/mdc lines; skip mgc/lov/lmv
# 351 UP osc snx11025-OST0007-osc-ffff8875ac1e7c00 3f30f170-90e6-b332-b141-a6d4a94a1829 5 10.100.100.12@o2ib1
#
# snx11035-OST0000_UUID 90767651352 54512631228 35277748388  61% /scratch2[OST:0]
#                             snx000-OST... tot     use     avail      00%    /scra[OST    :0     ]
_REX_OST_MAP = re.compile(r'^\s*(\d+)\s+(\S+)\s+(\S+)\s+(snx\d+-\S+)\s+(\S+)\s+(\d+)\s+(\S+@\S+)\s*$')
_REX_LFS_DF = re.compile(r'^\s*(snx\d+-\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+).\s+(\S+)\[([^:]+):(\d+)\]\s*$')

LCTL = 'lctl'
LCTL_DL_T = [LCTL, 'dl', '-t']
LFS = 'lfs'
LFS_DF = [LFS, 'df']

class LfsOstMap(SubprocessOutputDict):
    """
    Representation for the lctl dl -t command.  Generates a dict of form

        { file_system: { ost_name : { keys: values } } }

    This is a generally logical structure, although this map is always almost
    fed into a routine that tries to find multiple OSTs on the same OSS (i.e., a
    failover situation)
    """
    def __init__(self, *args, **kwargs):
        super(LfsOstMap, self).__init__(*args, **kwargs)
        self.subprocess_cmd = LCTL_DL_T
        self.load()

    def __repr__(self):
        """Serialize object into an ASCII string

        Returns a string that resembles the input used to initialize this object
        """
        repr_result = ""

        # Iterate over file systems within each time step
        for target_name in sorted(self.keys()):
            obd_data = self[target_name]
            for obd_name in sorted(list(obd_data.keys()), key=lambda x: int(obd_data[x]['index'])):
                keyvals = obd_data[obd_name]
                record_string = \
                    "%(index)3d %(status)2s %(role)3s %(role_id)s %(uuid)s %(ref_count)d %(nid)s\n" % keyvals
                repr_result += record_string

        return repr_result

    def load_str(self, input_str):
        """Parse the output of `lctl dl -t` to initialize self
        """
        degenerate_keys = 0
        for line in input_str.splitlines():
            match = _REX_OST_MAP.search(line)
            if match is not None:
                file_system, target_name = (match.group(4).split('-')[0:2])

                if file_system not in self:
                    self[file_system] = {}

                # Duplicates can happen if a file system is doubly mounted
                if target_name in self[file_system]:
                    degenerate_keys += 1

                self[file_system][target_name] = {
                    'index': int(match.group(1)),
                    'status': match.group(2).lower(),
                    'role': match.group(3).lower(),
                    'role_id': match.group(4),
                    'uuid': match.group(5),
                    'ref_count': int(match.group(6)),
                    'target_ip': match.group(7).split('@')[0],
                    'nid': match.group(7),
                }

    def get_failovers(self):
        """Identify OSSes with an abnormal number of OSTs

        Identify OSTs that are probably failed over and return a list of
        abnormal OSSes and the expected number of OSTs per OSS.
        """
        resulting_data = {}

        for file_system, ost_data in self.items():
            ost_counts = {} # key = ip address, val = ost count
            for ost_name, ost_values in ost_data.items():
                if ost_values['role'] != 'osc': # don't care about mdc, mgc
                    continue
                ip_addr = ost_values['target_ip']
                ost_counts[ip_addr] = ost_counts.get(ip_addr, 0) + 1

            # Get mode of OSTs per OSS to infer what "normal" OST/OSS ratio is
            histogram = {}
            for ip_addr, ost_count in ost_counts.items():
                if ost_count not in histogram:
                    histogram[ost_count] = 1
                else:
                    histogram[ost_count] += 1
            if not histogram:
                raise KeyError('no OSTs to count')
            mode = max(histogram, key=histogram.get)

            # Build a dict of { ip_addr: [ ostname1, ostname2, ... ], ... }
            abnormal_ips = {}
            for ost_name, ost_values in ost_data.items():
                if ost_values['role'] != 'osc': # don't care about mdc, mgc
                    continue
                ip_addr = ost_values['target_ip']
                if ost_counts[ip_addr] != mode:
                    if ip_addr in abnormal_ips:
                        abnormal_ips[ip_addr].append(ost_name)
                    else:
                        abnormal_ips[ip_addr] = [ost_name]

            resulting_data[file_system] = {
                'mode': mode,
                'abnormal_ips': abnormal_ips,
            }

        return resulting_data

class LfsOstFullness(SubprocessOutputDict):
    """
    Representation for the `lfs df` command.  Generates a dict of form

        { file_system: { ost_name : { keys: values } } }

    """
    def __init__(self, *args, **kwargs):
        super(LfsOstFullness, self).__init__(*args, **kwargs)
        self.subprocess_cmd = LFS_DF
        self.load()

    def __repr__(self):
        """Serialize object into an ASCII string

        Returns a string that resembles the input used to initialize this object:

            snx11025-OST0001_UUID 90767651352 63381521692 26424604184  71% /scratch1[OST:1]
        """
        repr_result = ""
        for target_name in sorted(self.keys()):
            obd_data = self[target_name]
            for obd_name in sorted(list(obd_data.keys()), key=lambda x: obd_data[x]['target_index']):
                keyvalues = obd_data[obd_name]
                repr_result += "%s-%s_UUID %ld %ld %ld %3d%% %s[%s:%d]\n" % (
                    target_name,
                    obd_name,
                    keyvalues['total_kib'],
                    keyvalues['used_kib'],
                    keyvalues['remaining_kib'],
                    # Note that lfs dl's percents are not divided by
                    # avail_kib, but rather the sum of used and remaining.
                    round(100.0 * keyvalues['used_kib'] / (keyvalues['remaining_kib'] + keyvalues['used_kib'])),
                    keyvalues['mount_pt'],
                    keyvalues['role'].upper(),
                    keyvalues['target_index'], )

        return repr_result

    def load_str(self, input_str):
        """Parse the output of `lfs df` to initialize self
        """
        degenerate_keys = 0
        for line in input_str.splitlines():
            match = _REX_LFS_DF.search(line)
            if match is not None:
                file_system, target_name = re.findall('[^-_]+', match.group(1))[0:2]

                if file_system not in self:
                    self[file_system] = {}

                # Duplicates can happen if a file system is doubly mounted
                if target_name in self[file_system]:
                    degenerate_keys += 1

                self[file_system][target_name] = {
                    'total_kib': int(match.group(2)),
                    'used_kib': int(match.group(3)),
                    'remaining_kib': int(match.group(4)),
                    'mount_pt': match.group(6),
                    'role': match.group(7).lower(),
                    'target_index': int(match.group(8)),
                }



#!/usr/bin/env python
"""
This connection provides an interface to the Cray XT/XC service database (SDB).
It is intended to be used to determine information about a node's configuration
within the network fabric to provide topological information.
"""

import sys
from . import common
from tokio.common import isstr

class CraySdbProc(common.SubprocessOutputDict):
    """Dictionary subclass that self-populates with Cray SDB data.

    Presents certain views of the Cray Service Database (SDB) as a
    dictionary-like object through the Cray SDB CLI.
    """
    def __init__(self, *args, **kwargs):
        """Load the processor configuration table from the SDB.

        Args:
            *args: Passed to tokio.connectors.common.SubprocessOutputDict
            **kwargs: Passed to tokio.connectors.common.SubprocessOutputDict
        """
        super(CraySdbProc, self).__init__(*args, **kwargs)
        self.subprocess_cmd = ['xtdb2proc', '-f', '-']
        self.key_order = [] # implement poor-man's OrderedDict
        self.load()

    def __repr__(self):
        """Serialize self in a format compatible with ``xtdb2proc``.

        Returns the object in the same format as the xtdb2proc output so that
        this object can be circularly serialized and deserialized.

        Returns:
        str: String representation of the processor mapping table in a
        format compatible with the output of ``xtdb2proc``.
        """
        repr_result = ""
        for _, record in self.items():
            line = []
            for key in self.key_order:
                try:
                    val = record[key]
                except KeyError:
                    sys.stderr.write("key does not appear in all records\n")
                    raise

                # Look at the type of each val and return the correponding string
                if val is None:
                    line.append("%s=null" % key)
                elif isstr(val):
                    line.append("%s='%s'" % (key, val))
                else:
                    line.append("%s=%s" % (key, val))

            repr_result += ','.join(line) + "\n"
        return repr_result

    def load_str(self, input_str):
        """Load the xtdb2proc data for a Cray system.

        Parses the xtdb2proc output text and inserts keys/values into self.

        Args:
            input_str (str): stdout of the ``xtdb2proc`` command
        """
        # Save the key order once
        check_keys = True
        for line in input_str.splitlines():
            if line.startswith('#') or line.strip() == "":
                continue
            fields = line.split(',')
            record = {}
            for field in fields:
                key, val = field.split('=', 1)
                val = val.strip().strip('\'"')
                # Look at the type of each val
                if val == "null":
                    val = None
                else:
                    try:
                        val = int(val)
                    except ValueError:
                        pass
                record[key] = val

                if check_keys:
                    self.key_order.append(key)
            check_keys = False
            key = int(record['processor_id'])
            assert 'processor_id' in record
            assert key not in self
            self.__setitem__(key, record)



"""
Retrieve the contents of an LMT database and cache it locally.
"""

import os
import datetime
import argparse
import tokio.connectors.lmtdb

def retrieve_tables(lmtdb, datetime_start, datetime_end, limit=None):
    """
    Given a start and end time, retrieve and cache all of the relevant contents
    of an LMT database.
    """

    min_ts_id, max_ts_id = lmtdb.get_ts_ids(datetime_start, datetime_end)

    for lmtdb_table, table_schema in tokio.connectors.lmtdb.LMTDB_TABLES.items():
        query_str = 'SELECT * from %s' % lmtdb_table

        ### if this table is indexed by time, restrict the time range.
        ### otherwise, dump the whole time-independent table
        if 'ts_id' in [x.lower() for x in table_schema['columns']]:
            query_str += " WHERE TS_ID >= %d AND TS_ID <= %d" % (min_ts_id, max_ts_id)

        ### limits (mostly for testing)
        if limit is not None:
            query_str += " LIMIT %d" % limit

        result = lmtdb.query(
            query_str=query_str,
            table=lmtdb_table,
            table_schema=table_schema)

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("start", type=str, help="start time in YYYY-MM-DDTHH:MM:SS format")
    parser.add_argument("end", type=str, help="end time in YYYY-MM-DDTHH:MM:SS format")
    parser.add_argument("-i", "--input", type=str, default=None, help="input cache db file")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    parser.add_argument("-l", "--limit", type=int, default=None,
                        help="restrict number of records returned per table")
    parser.add_argument("--host", type=str, default=None, help="database hostname")
    parser.add_argument("--user", type=str, default=None, help="database user")
    parser.add_argument("--password", type=str, default=None, help="database password")
    parser.add_argument("--database", type=str, default=None, help="database name")
    args = parser.parse_args(argv)

    start = datetime.datetime.strptime(args.start, "%Y-%m-%dT%H:%M:%S")
    end = datetime.datetime.strptime(args.end, "%Y-%m-%dT%H:%M:%S")

    cache_file = args.output

    ### Unlike nersc_jobsdb connector, LmtDb does not look for any
    ### environment variables intrinsically
    if args.input is not None:
        lmtdb = tokio.connectors.lmtdb.LmtDb(cache_file=args.input)
    else:
        lmtdb = tokio.connectors.lmtdb.LmtDb(
            dbhost=args.host,
            dbuser=args.user,
            dbpassword=args.password,
            dbname=args.database)

    retrieve_tables(lmtdb, start, end, args.limit)

    if cache_file is None:
        i = 0
        while True:
            cache_file = "lmtdb-%d.sqlite" % i
            if os.path.exists(cache_file):
                i += 1
            else:
                break
    print("Caching to %s" % cache_file)
    lmtdb.save_cache(cache_file)



#!/usr/bin/env python
"""
Various functions that may be of use in analyzing TOKIO data.  These are
provided as a convenience rather than a set of core functionality.
"""

### Subpackages to include in the tokio.analysis.* namespace
try:
    import tokio.analysis.umami
except ImportError:
    pass



#!/usr/bin/env python
"""
A higher-level interface that wraps various connectors and site-dependent
configuration to provide convenient abstractions upon which analysis tools can
be portably built.
"""

### Subpackages to include in the tokio.tools.* namespace
try:
    import tokio.tools.darshan
except ImportError:
    pass

try:
    import tokio.tools.hdf5
except ImportError:
    pass

try:
    import tokio.tools.topology
except ImportError:
    pass

try:
    import tokio.tools.lfsstatus
except ImportError:
    pass

### For backwards compatibility
try:
    import tokio.analysis.umami as umami
except ImportError:
    pass



"""
Provides CLI interfaces for
:meth:`tokio.connectors.slurm.Slurm.to_dataframe` and
:meth:`tokio.connectors.slurm.Slurm.to_json`.
"""

import os
import argparse
import tokio.connectors.slurm

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("slurm_jobid", type=str, help="slurm jobid to process")
    parser.add_argument("-n", "--native", action="store_true", default=True,
                        help="return output in native format")
    parser.add_argument("-j", "--json", action="store_true", help="return output in JSON format")
    parser.add_argument("-c", "--csv", action="store_true", help="return output in CSV format")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    args = parser.parse_args(argv)

    jobid = args.slurm_jobid

    # the following assumption only gets goofy if you have a file named the same
    # thing as a jobid
    if os.path.isfile(jobid):
        # treat the jobid as a cache file if it exists
        slurm_data = tokio.connectors.slurm.Slurm(cache_file=jobid)
    else:
        # treat the jobid as a slurm jobid and call sacct
        slurm_data = tokio.connectors.slurm.Slurm(jobid)

    # Serialize the object
    cache_file = args.output
    if cache_file is not None:
        print("Caching to %s" % cache_file)

    if cache_file is None:
        if args.csv:
            print((slurm_data.to_dataframe()).to_csv())
        elif args.json:
            print(slurm_data.to_json(indent=4, sort_keys=True))
        elif args.native:
            print(str(slurm_data))
        else:
            raise Exception("No output format specified")
    else:
        if args.csv:
            (slurm_data.to_dataframe()).to_csv(cache_file)
        elif args.json:
            with open(cache_file, 'w') as cache_fd:
                cache_fd.write(slurm_data.to_json())
        elif args.native:
            with open(cache_file, 'w') as cache_fd:
                cache_fd.write(str(slurm_data))
        else:
            raise Exception("No output format specified")



"""
Tools to find Darshan logs within a system-wide repository
"""

import os
import glob
from tokio.debug import debug_print
import tokio.tools.common
import tokio.tools.jobinfo
import tokio.connectors.darshan
import tokio.connectors.slurm

DARHSAN_LOG_NAME_STR = "%(username)s_%(exe)s_id%(jobid)s_%(month)s-%(day)s-%(seconds)s-%(logmod)s.darshan"
DARSHAN_LOG_GLOB_FIELDS = {
    "username": "*",
    "exe": "*",
    "jobid": "*",
    "month": "*",
    "day": "*",
    "seconds": "*",
    "logmod": "*",
}

def load_darshanlogs(datetime_start=None, datetime_end=None, username=None,
                     jobid=None, log_dir=None, system=None,
                     which=None, **kwargs):
    """Return parsed Darshan logs matching a set of criteria

    Finds Darshan logs that match the input criteria, loads them, and returns a
    dictionary of connectors.darshan.Darshan objects keyed by the full log file
    paths to the source logs.

    Args:
        datetime_start (datetime.datetime): date to begin looking for Darshan logs
        datetime_end (datetime.datetime): date to stop looking for Darshan logs
        username (str): username of user who generated the log
        jobid (int): jobid corresponding to Darshan log
        log_dir (str): path to Darshan log directory base
        system (str): key to pass to enumerate_dated_files's lookup_key
            when resolving darshan_log_dir
        which (str): 'base', 'total', and/or 'perf' as a comma-delimited string
        kwargs: arguments to pass to the connectors.darshan.Darshan object
            initializer

    Returns:
        dict: keyed by log file name whose values are connectors.darshan.Darshan
            objects

    Todo:
        * Use a default `darshan_log_dir` from `tokio.config`
    """
    if which is None:
        raise TypeError("which must be base, total, and/or perf")

    which_list = [x.lower() for x in which.split(',')]
    if 'base' not in which_list \
    and 'total' not in which_list \
    and 'perf' not in which_list:
        raise TypeError("which must be base, total, and/or perf")

    matching_logfiles = find_darshanlogs(datetime_start=datetime_start,
                                         datetime_end=datetime_end,
                                         username=username,
                                         jobid=jobid,
                                         system=system,
                                         log_dir=log_dir)

    results = {}
    for matching_logfile in matching_logfiles:
        results[matching_logfile] = tokio.connectors.darshan.Darshan(log_file=matching_logfile,
                                                                     **kwargs)
        if 'base' in which_list:
            results[matching_logfile].darshan_parser_base()
        if 'total' in which_list:
            results[matching_logfile].darshan_parser_total()
        if 'perf' in which_list:
            results[matching_logfile].darshan_parser_perf()

    return results

def find_darshanlogs(datetime_start=None, datetime_end=None, username=None, jobid=None,
                     log_dir=None, system=None):
    """Return darshan log file paths matching a set of criteria

    Attempts to find Darshan logs that match the input criteria.

    Args:
        datetime_start (datetime.datetime): date to begin looking for Darshan logs
        datetime_end (datetime.datetime): date to stop looking for Darshan logs
        username (str): username of user who generated the log
        jobid (int): jobid corresponding to Darshan log
        log_dir (str): path to Darshan log directory base
        system (str or None): key to pass to enumerate_dated_files's lookup_key
            when resolving darshan_log_dir

    Returns:
        list: paths of matching Darshan logs as strings

    Todo:
        * Abstract the jobid argument into a generic workload manager jobid tool
          that dispatches the correct workload manager connector.
        * Use a default `darshan_log_dir` from `tokio.config`
    """

    if datetime_start is None:
        if jobid is None:
            raise TypeError("datetime_start must be defined if jobid is not")
        else:
            datetime_start, _ = tokio.tools.jobinfo.get_job_startend(jobid=jobid)

    if datetime_end is None:
        datetime_end = datetime_start

    # first expand the eligible Darshan base directories
    if log_dir is None:
        log_dir = tokio.config.CONFIG.get('darshan_log_dirs')

    base_dirs = tokio.tools.common.enumerate_dated_files(start=datetime_start,
                                                         end=datetime_end,
                                                         template=log_dir,
                                                         lookup_key=system,
                                                         match_first=(system is not None))

    # then run another pass of enumerate_dated_files to resolve the path within
    # the base darshan log directory
    search_dirs = []
    for base_dir in base_dirs:
        search_dirs += tokio.tools.common.enumerate_dated_files(
            start=datetime_start,
            end=datetime_end,
            template=os.path.join(base_dir, "%-Y", "%-m", "%-d"),
            lookup_key=system)

    debug_print("Darshan log search directories are:\n  " + "\n  ".join(search_dirs))

    glob_fields = DARSHAN_LOG_GLOB_FIELDS.copy()
    if jobid is not None:
        glob_fields['jobid'] = jobid
    if username:
        glob_fields['username'] = username

    results = []
    for search_dir in search_dirs:
        results += glob.glob(os.path.join(search_dir, DARHSAN_LOG_NAME_STR % glob_fields))

    return list(set(results))



"""
Provides CLI interfaces into the :mod:`tokio.tools.lfsstatus` tool's
:meth:`tokio.tools.lfsstatus.get_failures` and
:meth:`tokio.tools.lfsstatus.get_fullness` methods.
"""

import argparse
import datetime
import json
import tokio.tools.lfsstatus as lfsstatus

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--fullness", nargs='?', const="", type=str,
                       help="path to lfs_df text file; summarize fullness of OSTs ")
    group.add_argument("--failure", nargs='?', const="", type=str,
                       help="path to ost_map text file; summarize failure state of OSSes and OSTs")
    parser.add_argument("-o", "--output", type=str, default=None, help="output file")
    parser.add_argument("filesystem", help="logical file system name (e.g., cscratch)")
    parser.add_argument("datetime", help="date and time of interest in YYYY-MM-DDTHH:MM:SS format")
    args = parser.parse_args(argv)

    target_datetime = datetime.datetime.strptime(args.datetime, "%Y-%m-%dT%H:%M:%S")
    if args.failure is not None:
        results = lfsstatus.get_failures(
            args.filesystem,
            target_datetime,
            cache_file=args.failure if args.failure != "" else None)
    elif args.fullness is not None:
        results = lfsstatus.get_fullness(
            args.filesystem,
            target_datetime,
            cache_file=args.fullness if args.fullness != "" else None)
    else:
        raise Exception('Neither --fullness nor --failure were specified')

    # Serialize the object
    cache_file = args.output
    if cache_file is None:
        print(json.dumps(results, indent=4, sort_keys=True))
    else:
        print("Caching to %s" % cache_file)
        json.dump(results, open(cache_file, 'w'))





#!/usr/bin/env python
"""Connect to Slurm via Slurm CLI outputs.

This connector provides Python bindings to retrieve information made available
through the standard Slurm saccount and scontrol CLI commands.
"""

import json
import errno
import warnings
import datetime
try:
    import StringIO as io
except ImportError:
    import io
import subprocess
import pandas
from tokio.common import isstr
from tokio.connectors.common import SubprocessOutputDict

SACCT = 'sacct'
SCONTROL = 'scontrol'

def expand_nodelist(node_string):
    """Expand Slurm compact nodelist into a set of nodes.

    Wraps ``scontrol show hostname nid0[5032-5159]`` to expand a Slurm nodelist
    string into a list of nodes.

    Args:
        node_string (str): Node list in Slurm's compact notation (e.g.,
            ``nid0[5032-5159]``)

    Returns:
        set: Set of strings which encode the fully expanded node names contained
        in `node_string`.
    """
    node_names = set([])
    try:
        output_str = subprocess.check_output([SCONTROL, 'show', 'hostname', node_string])
    except OSError as error:
        if error.errno == errno.ENOENT:
            raise type(error)(error.errno, "Slurm CLI (%s command) not found" % SCONTROL)
        raise

    if not isstr(output_str):
        output_str = output_str.decode() # for python3

    for line in output_str.splitlines():
        node_name = line.strip()
        if node_name:
            node_names.add(node_name)

    return node_names

def compact_nodelist(node_string):
    """Convert a string of nodes into compact representation.

    Wraps ``scontrol show hostlist nid05032,nid05033,...`` to compress a list of
    nodes to a Slurm nodelist string.  This is effectively the reverse of
    ``expand_nodelist()``

    Args:
        node_string (str): Comma-separated list of node names (e.g.,
            ``nid05032,nid05033,...``)
    Returns:
        str: The compact representation of `node_string` (e.g.,
        ``nid0[5032-5159]``)
    """
    if not isstr(node_string):
        node_string = ','.join(list(node_string))

    try:
        node_string = subprocess.check_output([SCONTROL, 'show', 'hostlist', node_string]).strip()
    except OSError as error:
        if error.errno == errno.ENOENT:
            # "No such file or directory" from subprocess.check_output
            pass

    if isstr(node_string):
        return node_string
    else:
        return node_string.decode()

_RECAST_KEY_MAP = {
    'start':    (
        lambda x: datetime.datetime.strptime(x, "%Y-%m-%dT%H:%M:%S"),
        lambda x: x.strftime("%Y-%m-%dT%H:%M:%S")
    ),
    'end':      (
        lambda x: datetime.datetime.strptime(x, "%Y-%m-%dT%H:%M:%S"),
        lambda x: x.strftime("%Y-%m-%dT%H:%M:%S")
    ),
    'nodelist': (
        expand_nodelist,
        compact_nodelist
    ),
}
"""dict: Methods to convert Slurm string outputs into Python objects

This table provides the methods to apply to various Slurm output keys to convert
them from strings (the default Slurm output type) into more useful Python
objects such as datetimes or lists.

* ``value[0]`` is the function to cast to Python
* ``value[1]`` is the function to cast back to a string
"""

class SlurmEncoder(json.JSONEncoder):
    """Encode sets as lists and datetimes as ISO 8601.
    """
    def default(self, o): # pylint: disable=E0202
        if isinstance(o, set):
            return list(o)
        elif isinstance(o, datetime.datetime):
            return o.strftime("%Y-%m-%dT%H:%M:%S")
        return json.JSONEncoder.default(self, o)

class Slurm(SubprocessOutputDict):
    """Dictionary subclass that self-populates with Slurm output data

    Presents a schema that is keyed as::

        {
            taskid: {
                slurmfield1: value1
                slurmfield2: value2
                ...
            }
        }

    where taskid can be any of

        * jobid
        * jobid.<step>
        * jobid.batch
    """
    def __init__(self, jobid=None, *args, **kwargs):
        """Load basic information from Slurm

        Args:
            jobid (str): Slurm Job ID associated with data this object contains

        Attributes:
            jobid (str): Slurm Job ID associated with data contained in this
                object
        """
        super(Slurm, self).__init__(*args, **kwargs)
        self.subprocess_cmd = [SACCT]
        if jobid is not None:
            self.jobid = str(jobid)
        else:
            self.jobid = jobid
        self.load()

    def __repr__(self):
        """Serialize object in the same format as sacct.

        Returns:
            str: Serialized version of self in a similar format as the ``sacct``
            output so that this object can be circularly serialized and
            deserialized.
        """
        output_str = ""
        key_order = ['jobidraw']
        for counters in self.values():
            # print the column headers on the first pass
            if output_str == "":
                for key in counters:
                    if key not in key_order:
                        key_order.append(key)
                output_str = '|'.join(key_order)

            print_values = []
            for key in key_order:
                value = counters[key]
                # convert specific keys back into strings
                if key in _RECAST_KEY_MAP:
                    value = _RECAST_KEY_MAP[key][1](value)
                print_values.append(value)
            output_str += "\n" + '|'.join(print_values)
        return output_str

    def load(self):
        """Initialize values either from cache or sacct
        """
        if self.from_string is not None:
            self.load_str(self.from_string)
        elif self.cache_file:
            self.load_cache()
        elif self.jobid is None:
            raise Exception("either jobid or cache_file must be specified on init")
        else:
            self.load_keys('jobidraw', 'start', 'end', 'nodelist')

    def load_keys(self, *keys):
        """Retrieve a list of keys from sacct and insert them into self.

        This always invokes sacct and can be used to overwrite the contents of a
        cache file.

        Args:
            *keys (list): Slurm attributes to include; names should be
                valid input to `sacct --format` CLI utility.
        """
        if self.jobid is None:
            raise Exception("Slurm.jobid is None")

        args = ['--jobs', self.jobid,
                '--format=%s' % ','.join(keys),
                '--parsable2']

        self._load_subprocess(*args)

        return self

    def load_str(self, input_str):
        """Load from either a json cache or the output of sacct
        """
        loaded_data = None
        try:
            loaded_data = self.from_json(input_str)
        except ValueError:
            pass

        if loaded_data is None:
            loaded_data = parse_sacct(input_str)

        self.update(loaded_data)

        self._recast_keys()

    def _recast_keys(self, *target_keys):
        """Convert own keys into native Python objects.

        Scan self and convert special keys into native Python objects where
        appropriate.  If no keys are given, scan everything.  Do NOT attempt
        to recast anything that is not a string--this is to avoid relying on
        expand_nodelist if a key is already recast since expand_nodelist does
        not function outside of an environment containing Slurm.

        Args:
            *target_keys (list, optional): Only convert these keys into native
                Python object types.  If omitted, convert all keys.
        """
        scan_keys = len(target_keys)
        for counters in self.values():
            # if specific keys were passed, only look for those keys
            if scan_keys > 0:
                for key in target_keys:
                    value = counters[key]
                    if key in _RECAST_KEY_MAP and isstr(value):
                        counters[key] = _RECAST_KEY_MAP[key][0](value)
            # otherwise, attempt to recast every key
            else:
                for key, value in counters.items():
                    if key in _RECAST_KEY_MAP and isstr(value):
                        counters[key] = _RECAST_KEY_MAP[key][0](value)

#   def get_task_startend(self, taskid=self.jobid):
#       """
#       Convert raw Slurm keys into datetime objects and return them
#       """
#       # initialize self if needed
#       if len(self) == 0 or taskid not in self.keys():
#           self.load_keys('jobidraw', 'start', 'end')
#
#       # make sure the requested taskid exists
#       if taskid not in self.keys():
#           raise Exception("Invalid task id '%s' for job '%s'" % (str(taskid), str(self.jobid)))
#
#       # if somehow got partially loaded
#       if 'start' not in self[taskid] or 'end' not in self[taskid]:
#           raise Exception("No start/end information for task id %s" % taskid)
#
#       return self[taskid]['start'], self[taskid]['end']

    def get_job_nodes(self):
        """Return a list of all job nodes used.

        Creates a list of all nodes used across all tasks for the self.jobid.
        Useful if the object contains only a subset of tasks executed by the
        Slurm job.

        Returns:
            set: Set of node names used by the job described by this object
        """
        nodelist = set([])

        for counters in self.values():
            for jobnode in counters['nodelist']:
                nodelist.add(jobnode)

        return nodelist

    def get_job_startend(self):
        """Find earliest start and latest end time for a job.

        For an entire job and all its tasks, find the absolute earliest start
        time and absolute latest end time.

        Returns:
            tuple: Two-item tuple of (earliest start time, latest end time) in
                whatever type ``self['start']`` and ``self['end']`` are stored
        """
        min_start = None
        max_end = None
        for counters in self.values():
            if min_start is None or min_start > counters['start']:
                min_start = counters['start']
            if max_end is None or max_end < counters['end']:
                max_end = counters['end']

        return min_start, max_end

    def get_job_ids(self):
        """Return the top-level jobid(s) contained in object.

        Retrieve the jobid(s) contained in self without any accompanying taskid
        information.

        Returns:
            list of str: list of jobid(s) contained in self.
        """
        jobids = []
        for rawjobid in self:
            if '.' not in rawjobid:
                jobids.append(rawjobid)
        return jobids

    def to_json(self, **kwargs):
        """Return a json-encoded string representation of self.

        Serializes self to json using ``_RECAST_KEY_MAP`` to convert Python
        types back into JSON-compatible types.

        Returns:
            str: JSON representation of self
        """
        return json.dumps(self, cls=SlurmEncoder, **kwargs)

    def from_json(self, json_string):
        """Initialize self from a JSON-encoded string.

        Args:
            json_string (str): JSON representation of self
        """
        decoded_dict = json.loads(json_string)
        for key, value in decoded_dict.items():
            self[key] = value
        self._recast_keys()

    def to_dataframe(self):
        """Convert self into a Pandas DataFrame.

        Returns a Pandas DataFrame representation of this object.

        Returns:
            pandas.DataFrame: DataFrame representation of the same schema as
            the Slurm ``sacct`` command.
        """
        buf = io.StringIO(str(self))
        return pandas.read_csv(buf, sep='|', parse_dates=['start', 'end'])

def parse_sacct(sacct_str):
    """Convert output of ``sacct -p`` into a dictionary.

    Parses the output of ``sacct -p`` and return a dictionary with the full (raw)
    contents.

    Args:
        sacct_str (str): stdout of an invocation of ``sacct -p``

    Returns:
        dict: Keyed by Slurm Job ID and whose values are dicts containing
        key-value pairs corresponding to the Slurm quantities returned
        by ``sacct -p``.
    """
    result = {}
    cols = []
    for lineno, line in enumerate(sacct_str.splitlines()):
        fields = line.split('|')
        if lineno == 0:
            cols = [x.lower() for x in fields]
        else:
            record = {}
            jobidraw = fields[0]
            if jobidraw in result:
                warnings.warn("Duplicate raw jobid '%s' found" % jobidraw)
            for col, key in enumerate(cols):
                record[key] = fields[col]
            result[jobidraw] = record
    return result



#!/usr/bin/env python
"""
Extract job info from the NERSC jobs database.  Accessing the MySQL database is
not required (i.e., if you have everything stored in a cache, you never have to
touch MySQL).  However if you do have to connect to MySQL, you must set the
following environment variables::

    NERSC_JOBSDB_HOST
    NERSC_JOBSDB_USER
    NERSC_JOBSDB_PASSWORD
    NERSC_JOBSDB_DB

If you do not know what to use as these credentials, you will have to rely on
a cache database.
"""

import os
import warnings
import tokio.connectors.cachingdb as cachingdb

NERSC_JOBSDB_SCHEMA = {
    'columns': ['STEPID', 'HOSTNAME', 'START', 'COMPLETION', 'NUMNODES'],
    'primary_key': ['STEPID'],
}
NERSC_JOBSDB_QUERY = """
SELECT
    s.stepid,
    s.hostname,
    s.start,
    s.completion,
    s.numnodes
FROM
    summary AS s
"""

HIT_MEMORY = 0
HIT_CACHE_DB = 1
HIT_REMOTE_DB = 2

class NerscJobsDb(cachingdb.CachingDb):
    """
    Connect to and interact with the NERSC jobs database.  Maintains a query
    cache where the results of queries are cached in memory.  If a query is
    repeated, its values are simply regurgitated from here rather than touching
    any databases.

    If this class is instantiated with a cache_file argument, all queries will
    go to that SQLite-based cache database if they are not found in the
    in-memory cache.

    If this class is not instantiated with a cache_file argument, all queries
    that do not exist in the in-memory cache will go out to the MySQL database.

    The in-memory query caching is possible because the job data in the NERSC
    jobs database is immutable and can be cached indefinitely once it appears
    there.  At any time the memory cache can be committed to a cache database to
    be used or transported later.
    """
    def __init__(self, dbhost=None, dbuser=None, dbpassword=None, dbname=None, cache_file=None):
        # in-memory query cache
        self.cached_queries = {}
        self.last_results = None # for debugging

        if dbhost is None:
            dbhost = os.environ.get('NERSC_JOBSDB_HOST')
        if dbuser is None:
            dbuser = os.environ.get('NERSC_JOBSDB_USER')
        if dbpassword is None:
            dbpassword = os.environ.get('NERSC_JOBSDB_PASSWORD')
        if dbname is None:
            dbname = os.environ.get('NERSC_JOBSDB_DB')

        super(NerscJobsDb, self).__init__(
            dbhost=dbhost,
            dbuser=dbuser,
            dbpassword=dbpassword,
            dbname=dbname,
            cache_file=cache_file)

    def drop_cache(self):
        """
        Clear the query cache
        """
        super(NerscJobsDb, self).drop_cache()
        self.cached_queries = {}

    def query(self, query_str, query_variables=(), nocache=False):
        """
        Pass a query through all layers of cache and return on the first hit.
        """

        ### Collapse query string to remove extraneous whitespace
        query_str = ' '.join(query_str.split())
        cache_key = query_str % {'ps': '%s'} % query_variables

        ### FIRST: check cache for the results of this query--we can get away
        ### with this because the NERSC Jobs DB is append-only
        if cache_key in self.cached_queries and not nocache:
            index_start, index_end = self.cached_queries[cache_key]
            results = self.saved_results['summary']['rows'][index_start:index_end]
            self.last_hit = HIT_MEMORY
        else:
            if 'summary' not in self.saved_results:
                index_start = 0
            else:
                index_start = len(self.saved_results['summary']['rows'])
            results = super(NerscJobsDb, self).query(
                query_str,
                query_variables,
                table='summary',
                table_schema=NERSC_JOBSDB_SCHEMA)
            index_end = len(self.saved_results['summary']['rows'])
            if not nocache:
                self.cached_queries[cache_key] = (index_start, index_end)

        self.last_results = results # for debugging
        return results

    def get_concurrent_jobs(self, start_timestamp, end_timestamp, nersc_host):
        """
        Grab all of the jobs that were running, in part or in full, during the
        time window bounded by start_timestamp and end_timestamp.  Then
        calculate the fraction overlap for each job to calculate the number of
        core hours that were burned overall during the start/end time of
        interest.
        """
        query_str = NERSC_JOBSDB_QUERY + """
        WHERE
            s.hostname = %(ps)s
        AND s.completion > %(ps)s
        AND s.start < %(ps)s
        """

        totals = {
            'nodehrs': 0.0,
            'numnodes': 0,
            'numjobs': 0,
        }

        results = self.query(
            query_str,
            (nersc_host, start_timestamp, end_timestamp))

        for (_, _, this_start, this_end, nnodes) in results:
            real_start = max(this_start, start_timestamp)
            real_end = min(this_end, end_timestamp)
            delta_t = real_end - real_start
            totals['nodehrs'] += nnodes * delta_t
            totals['numnodes'] += nnodes
            totals['numjobs'] += 1

        totals['nodehrs'] /= 3600.0

        return totals

#   def get_concurrent_jobs(self, start_timestamp, end_timestamp, nersc_host):
    def get_job_startend(self, jobid, nersc_host):
        """Return start and end time for a given job id

        Retrieves the time a job started and completed.

        Args:
            jobid (str): Job ID of interest
            nersc_host (str): NERSC host to which job ID of interest maps

        Returns:
            tuple of datetime.datetime: Two-item tuple of (start time,
            end time)
        """
        query_str = NERSC_JOBSDB_QUERY + """
        WHERE
            s.stepid LIKE %(ps)s
        AND s.hostname = %(ps)s
        ORDER BY s.completion
        """

        results = self.query(query_str, query_variables=("%s.%%" % jobid, nersc_host))

        start = None
        end = None
        for (_, _, this_start, this_end, _) in results:
            if start is not None:
                warnings.warn("Multiple start+end times found for job %s" % jobid)
            start = this_start
            end = this_end
        return (start, end)



#!/usr/bin/env python
"""
Given a file system and a datetime, return summary statistics about the OST
fullness at that time
"""

import time
import datetime
import tokio.tools.common
import tokio.config
import tokio.connectors.nersc_lfsstate as nersc_lfsstate

DEFAULT_FULLNESS_PROVIDERS = ['hdf5', 'nersc_lfsstate']

def get_fullness(file_system, datetime_target, **kwargs):
    """Get file system fullness

    Is a convenience wrapper for `get_summary`.

    Args:
        file_system (str): Logical name of file system whose data should be
            retrieved (e.g., cscratch)
        datetime_target (datetime.datetime): Time at which requested data
            should be retrieved

    Returns:
        dict: various statistics about the file system fullness

    Raises:
        tokio.ConfigError: When no valid providers are found
    """
    providers = tokio.config.CONFIG.get('lfsstatus_fullness_providers', DEFAULT_FULLNESS_PROVIDERS)
    match = False
    fullness = {}
    for provider in providers:
        if provider == 'hdf5':
            match = True
            try:
                fullness = get_fullness_hdf5(file_system, datetime_target)
            except KeyError:
                # get_fullness_hdf5 throws KeyError if fullness data is not available
                match = False
            if fullness:
                return fullness
        if provider == 'nersc_lfsstate':
            match = True
            fsname = tokio.config.CONFIG.get('fsname_to_backend_name', {}).get(file_system)
            try:
                fullness = get_lfsstate(fsname if fsname else file_system,
                                        datetime_target,
                                        "fullness",
                                        **kwargs)
            except IOError:
                # get_lfsstate raises IOError if fullness data is not available
                match = False
            if fullness:
                return fullness

    if match:
        return fullness

    raise tokio.ConfigError("No valid lfsstatus fullness providers found")


def get_failures(file_system, datetime_target, **kwargs):
    """Get file system failures

    Is a convenience wrapper for `get_summary`.

    Args:
        file_system (str): Logical name of file system whose data should be
            retrieved (e.g., cscratch)
        datetime_target (datetime.datetime): Time at which requested data
            should be retrieved
        cache_file (str): Basename of file to search for the requested data

    Returns:
        dict: various statistics about the file system fullness
    """
    fsname = tokio.config.CONFIG.get('fsname_to_backend_name', {}).get(file_system)
    return get_lfsstate(fsname if fsname else file_system, datetime_target, "failures", **kwargs)


def get_fullness_lfsstate(file_system, datetime_target, cache_file=None):
    """Get file system fullness from nersc_lfsstate connector

    Wrapper around the generic get_lfsstate function.

    Args:
        file_system (str): Lustre file system name of the file system whose
            data should be retrieved (e.g., snx11025)
        datetime_target (datetime.datetime): Time at which requested data
            should be retrieved
        cache_file (str): Basename of file to search for the requested data

    Returns:
        Whatever is returned by :func:`tokio.tools.lfsstatus.get_lfsstate`
    """
    return get_lfsstate(file_system, datetime_target, 'fullness', cache_file)

def get_failures_lfsstate(file_system, datetime_target, cache_file=None):
    """Get file system failures from nersc_lfsstate connector

    Wrapper around the generic get_lfsstate function.

    Args:
        file_system (str): Lustre file system name of the file system whose
            data should be retrieved (e.g., snx11025)
        datetime_target (datetime.datetime): Time at which requested data
            should be retrieved
        cache_file (str): Basename of file to search for the requested data

    Returns:
        Whatever is returned by get_lfsstate
    """
    return get_lfsstate(file_system, datetime_target, 'failures', cache_file)

def get_lfsstate(file_system, datetime_target, metric, cache_file=None):
    """Get file system fullness or failures

    Given a file system name (e.g., snx11168) and a datetime object

        1. locate and load the lfs-df (fullness) or ost map (failures) file
        2. find the sample immediately preceding the datetime (don't find one
           that overlaps it
        3. return summary statistics about the OST fullness or OST failures

    Args:
        file_system (str): Lustre file system name of the file system whose
            data should be retrieved (e.g., snx11025)
        datetime_target (datetime.datetime): Time at which requested data
            should be retrieved
        metric (str): either "fullness" or "failures"
        cache_file (str): Basename of file to search for the requested data

    Returns:
        dict: various statistics about the file system fullness

    Raises:
        ValueError: if ``metric`` does not contain a valid option
        IOError: when no valid data sources can be found for the given date
    """
    if metric == "fullness":
        template_path = tokio.config.CONFIG['lfsstatus_fullness_files']
    elif metric == "failures":
        template_path = tokio.config.CONFIG['lfsstatus_map_files']
    else:
        raise ValueError("unknown metric " + metric)

    if cache_file is None:
        # We assume a 1 day lookbehind.  Very wasteful, but we index on dates in
        # local time using GMT-based unix times so we often need to look back to
        # the previous day's index.  The lookahead can be much more conservative
        # since it only needs to compensate for sampling intervals (15 min in
        # practice at NERSC)
        ost_health_files = tokio.tools.common.enumerate_dated_files(
            start=datetime_target - datetime.timedelta(days=1),
            end=datetime_target + datetime.timedelta(hours=1),
            template=template_path,
            match_first=True)

        for index, df_file in enumerate(ost_health_files):
            ost_health_files[index] = ost_health_files[index]
    else:
        ost_health_files = [cache_file]

    if not ost_health_files:
        raise IOError("No OST health files found in %s for %s" % (
            template_path,
            str(datetime_target)))

    # We can get away with the following because NerscLfsOstFullness,
    # NerscLfsOstMap, and NerscLfsOstMap.get_failovers all have the same
    # structure
    if metric == "fullness":
        ost_health = None
        for df_file in ost_health_files:
            if ost_health is None:
                ost_health = nersc_lfsstate.NerscLfsOstFullness(cache_file=df_file)
            else:
                ost_health.update(nersc_lfsstate.NerscLfsOstFullness(cache_file=df_file))
    elif metric == "failures":
        ost_map = None
        for map_file in ost_health_files:
            if ost_map is None:
                ost_map = nersc_lfsstate.NerscLfsOstMap(cache_file=map_file)
            else:
                ost_map.update(nersc_lfsstate.NerscLfsOstMap(cache_file=map_file))
        ost_health = ost_map.get_failovers()

    timestamps = sorted([int(x) for x in ost_health])

    # Unoptimized walk through to find our timestamp of interest
    target_timestamp = int(time.mktime(datetime_target.timetuple()))
    # If the day's records start after the target time stamp, just report the
    # first record (target_index=0)
    target_index = 0
    for index, timestamp in enumerate(timestamps):
        if timestamp >= target_timestamp:
            if index == 0:
                break
            else:
                target_index = index - 1
                break

    fs_data = ost_health[timestamps[target_index]][file_system]

    if metric == "fullness":
        results = _summarize_fullness(fs_data)
    if metric == "failures":
        results = _summarize_failover(fs_data)

    # In case you want to interpolate--hope is that we have enough data points
    # where OST volumes will not change significantly enough to require
    # interpolation
    if target_index < (len(timestamps) - 1):
        results['ost_next_timestamp'] = timestamps[target_index] + 1
    results.update({
        'ost_actual_timestamp': timestamps[target_index],
        'ost_requested_timestamp': target_timestamp,
    })
    return results


def get_fullness_hdf5(file_system, datetime_target):
    """Get file system fullness from an HDF5 object

    Given a file system name (e.g., snx11168) and a datetime object,
    return summary statistics about the OST fullness.

    Args:
        file_system (str): Name of file system whose data should be retrieved
        datetime_target (datetime.datetime): Time at which requested data
            should be retrieved

    Returns:
        dict: various statistics about the file system fullness

    Raises:
        ValueError: if an OST name is encountered which does not conform to
            a naming convention from which an OST index can be derived
    """
    # For concordance with the lfsstate version, assume a generous lookbehind
    # that captures at least a few timesteps.  Lookahead does not need to be as
    # generous.
    df_bytes = tokio.tools.hdf5.get_dataframe_from_time_range(
        fsname=file_system,
        dataset_name='/fullness/bytes',
        datetime_start=datetime_target - datetime.timedelta(hours=1),
        datetime_end=datetime_target + datetime.timedelta(hours=1))
    df_bytes_tot = tokio.tools.hdf5.get_dataframe_from_time_range(
        fsname=file_system,
        dataset_name='/fullness/bytestotal',
        datetime_start=datetime_target - datetime.timedelta(hours=1),
        datetime_end=datetime_target + datetime.timedelta(hours=1))

    # Bail if nothing is found
    if df_bytes is None or df_bytes_tot is None or len(df_bytes) == 0 or len(df_bytes_tot) == 0:
        return {}

    # NAs must be filled or else fullness/total values will be spuriously low
    df_bytes = df_bytes.fillna(method='ffill').fillna(method='bfill')
    df_bytes_tot = df_bytes_tot.fillna(method='ffill').fillna(method='bfill')

    # Find closest matching timestamp
    iloc = df_bytes.index.get_loc(datetime_target, method='nearest')

    # Build a dictionary that _summarize_fullness will accept as input
    results = {}
    for ostname, value in df_bytes.iloc[iloc].items():
        ostname_key = ostname.split('-')[-1]
        try:
            target_index = int(ostname_key.lower().lstrip('ost'), 16)
        except ValueError as error:
            raise type(error)("Cannot derive OST index from name '%s'" % ostname)
        used_kib = int(value / 1024.0)
        total_kib = int(df_bytes_tot.iloc[iloc][ostname] / 1024.0)
        results[ostname_key] = {
            'used_kib': used_kib,
            'total_kib': total_kib,
            'target_index': target_index,
        }

    summarized_data = _summarize_fullness(results)

    # for interpolation and error bounding
    if iloc < len(df_bytes.index):
        summarized_data['ost_next_timestamp'] = df_bytes.index[iloc + 1].to_pydatetime()

    summarized_data.update({
        'ost_actual_timestamp': df_bytes.index[iloc].to_pydatetime(),
        'ost_requested_timestamp': datetime_target,
    })

    return summarized_data


def _summarize_fullness(fs_data):
    """Summarize fullness data for a single time record

    Given an fs_data dict, generate a dict of summary statistics.  Expects
    fs_data dict of form generated by `nersc_lfsstate.NerscLfsOstFullness`::

        {
             "MDT0000": {
                 "mount_pt": "/scratch1",
                 "remaining_kib": 2147035984,
                 "target_index": 0,
                 "total_kib": 2255453580,
                 "used_kib": 74137712
             },
             "OST0000": {
                 "mount_pt": "/scratch1",
                 "remaining_kib": 28898576320,
                 "target_index": 0,
                 "total_kib": 90767651352,
                 "used_kib": 60894630700
             },
             ...
        }

    Args:
        fs_data (dict): a single timestamp and file system record taken from
            a nersc_lfsstate.NerscLfsOstFullness object

    Returns:
        dict: summary metrics about the state of the file system fullness
    """
    results = {
        'ost_least_full_kib': None,
        'ost_most_full_kib': 0,
        'ost_avg_full_kib': 0,
        'ost_avg_full_pct': 0,
        'ost_count': 0,
    }

    for ost_name, ost_data in fs_data.items():
        # Only care about OSTs, not MDTs or MGTs
        if not ost_name.lower().startswith('ost'):
            continue
        results['ost_count'] += 1
        results['ost_avg_full_kib'] += ost_data['used_kib']
        results['ost_avg_full_pct'] += ost_data['total_kib']
        if results['ost_least_full_kib'] is None \
        or results['ost_least_full_kib'] > ost_data['used_kib']:
            results['ost_least_full_kib'] = ost_data['used_kib']
            results['ost_least_full_name'] = ost_name
            results['ost_least_full_pct'] = 100.0 * ost_data['used_kib'] / ost_data['total_kib']

        if results['ost_most_full_kib'] < ost_data['used_kib']:
            results['ost_most_full_kib'] = ost_data['used_kib']
            results['ost_most_full_name'] = ost_name
            results['ost_most_full_pct'] = 100.0 * ost_data['used_kib'] / ost_data['total_kib']


    # If there are no osts, this will break
    try:
        results['ost_avg_full_kib'] = int(float(results['ost_avg_full_kib']) \
                                          / float(results['ost_count']))
        results['ost_avg_full_pct'] = 100.0 * float(results['ost_avg_full_kib']) \
                                      / float(results['ost_avg_full_pct']) \
                                      * float(results['ost_count'])
    except ZeroDivisionError:
        pass

    results['ost_least_full_id'] = fs_data[results['ost_least_full_name']]['target_index']
    results['ost_most_full_id'] = fs_data[results['ost_most_full_name']]['target_index']
    return results


def _summarize_failover(fs_data):
    """Summarize failover data for a single time record

    Given an fs_data dict, generate a dict of summary statistics. Expects
    fs_data dict of the form generated by parse_lustre_txt.get_failovers::

        {
            "abnormal_ips": {
                "10.100.104.140": [
                    "OST0087",
                    "OST0086",
                    ...
                ],
                "10.100.104.43": [
                    "OST0025",
                    "OST0024",
                    ...
                ]
            },
            "mode": 1
        }

    Args:
        fs_data (dict): a single timestamp and file system record taken from
            the output of `nersc_lfsstate.NerscLfsOstMap.get_failovers`

    Returns:
        dict: summary metrics about the state of failovers on the file system
    """
    num_abnormal_ip = len(fs_data['abnormal_ips'])
    num_abnormal_osts = 0
    if num_abnormal_ip:
        for _, ost_list in fs_data['abnormal_ips'].items():
            num_abnormal_osts += len(ost_list)
        avg_overload = float(num_abnormal_osts) / float(num_abnormal_ip)
        avg_overload_factor = avg_overload / float(fs_data['mode'])
    else:
        avg_overload = 0.0
        avg_overload_factor = 1.0

    return {
        'ost_overloaded_oss_count': num_abnormal_ip,
        'ost_overloaded_ost_count': num_abnormal_osts,
        'ost_avg_overloaded_ost_per_oss': avg_overload,
        'ost_avg_overloaded_overload_factor': avg_overload_factor,
    }



"""
Common methods and classes used by connectors
"""

import os
import sys
import gzip
import json
import errno
import tarfile
import warnings
import mimetypes
import subprocess
from tokio.common import isstr

class CacheableDict(dict):
    """Generic class to support connectors that are dicts that can be cached as JSON

    When deriving from this class, the child object will have to define its own
    ``load_native()`` method to be invoked when ``input_file`` is not JSON.

    """
    def __init__(self, input_file=None):
        """Either initialize as empty or load from cache

        Args:
            input_file (str): Path to either a JSON file representing the dict
                or a native file that will be parsed into a JSON-compatible
                format
        """
        super(CacheableDict, self).__init__()
        self.input_file = input_file
        self.load()

    def load(self, input_file=None):
        """Wrapper around the filetype-specific loader.

        Infer the type of input being given, dispatch the correct loading
        function, and populate keys/values.

        Args:
            input_file (str or None): The input file to load.  If not specified,
                uses whatever self.input_file is
        """
        if input_file:
            self.input_file = input_file

        if self.input_file is None:
            return

        if not os.path.exists(self.input_file):
            raise OSError("Input file %s does not exist" % self.input_file)

        try:
            self.load_json()
        except ValueError:
            self.load_native()

    def load_native(self, input_file=None):
        """Parse an uncached, native object

        This is a stub that should be overloaded on derived classes.

        Args:
            input_file (str or None): The input file to load.  If not specified,
                uses whatever self.input_file is
        """
        pass

    def load_json(self, input_file=None):
        """Loads input from serialized JSON

        Load the serialized format of this object, encoded as a json dictionary.
        This is the converse of the save_cache() method.

        Args:
            input_file (str or None): The input file to load.  If not specified,
                uses whatever self.input_file is
        """
        if input_file:
            self.input_file = input_file

        _, encoding = mimetypes.guess_type(self.input_file)

        if encoding == 'gzip':
            open_func = gzip.open
        else:
            open_func = open

        for key, val in json.load(open_func(self.input_file, 'r')).items():
            self.__setitem__(key, val)

    def save_cache(self, output_file=None, **kwargs):
        """Serializes self into a JSON output.

        Save the dictionary in a JSON file.  This output can be read back in using
        load_json().

        Args:
            output_file (str or None): Path to file to which json should be
                written.  If None, write to stdout.  Default is None.
            kwargs (dict): Additional arguments to be passed to json.dumps()
        """
        if output_file is None:
            self._save_cache(sys.stdout, **kwargs)
        else:
            with open(output_file, 'w') as output:
                self._save_cache(output, **kwargs)

    def _save_cache(self, output, **kwargs):
        """Generates serialized representation of self

        Args:
            output: Object with a ``.write()`` method into which the serialized
                form of self will be passed
        """
        output.write(json.dumps(self, **kwargs))

class SubprocessOutputDict(dict):
    """Generic class to support connectors that parse the output of a subprocess

    When deriving from this class, the child object will have to

        1. Define subprocess_cmd after initializing this parent object
        2. Define self.__repr__ (if necessary)
        3. Define its own self.load_str
        4. Define any introspective analysis methods

    """
    def __init__(self, cache_file=None, from_string=None, silent_errors=False):
        super(SubprocessOutputDict, self).__init__(self)
        self.cache_file = cache_file
        self.silent_errors = silent_errors
        self.from_string = from_string
        self.subprocess_cmd = []

    def load(self, cache_file=None):
        """Load based on initialization state of object

        Args:
            cache_file (str or None): The cached input file to load.  If not
                specified, uses whatever self.cache_file is
        """
        if cache_file:
            self.cache_file = cache_file

        if self.from_string is not None:
            self.load_str(self.from_string)
        elif self.cache_file:
            self.load_cache()
        elif self.subprocess_cmd:
            self._load_subprocess()

    def _load_subprocess(self, *args):
        """Run a subprocess and pass its stdout to a self-initializing parser
        """

        cmd = self.subprocess_cmd
        if args:
            cmd += args

        try:
            if self.silent_errors:
                with open(os.devnull, 'w') as devnull:
                    output_str = subprocess.check_output(cmd, stderr=devnull)
            else:
                output_str = subprocess.check_output(cmd)
        except subprocess.CalledProcessError as error:
            warnings.warn("%s returned nonzero exit code (%d)" % (cmd, error.returncode))
            output_str = error.output
        except OSError as error:
            if error.errno == errno.ENOENT:
                raise type(error)(error.errno, "%s command not found" % self.subprocess_cmd[0])
            raise

        if isstr(output_str):
            # Python 2 - subprocess.check_output returns a string
            self.load_str(output_str)
        else:
            # Python 3 - subprocess.check_output returns encoded bytes
            self.load_str(output_str.decode())

    def load_cache(self, cache_file=None):
        """Load subprocess output from a cached text file

        Args:
            cache_file (str or None): The cached input file to load.  If not
                specified, uses whatever self.cache_file is
        """
        if cache_file:
            self.cache_file = cache_file

        _, encoding = mimetypes.guess_type(self.cache_file)
        if encoding == 'gzip':
            input_fp = gzip.open(self.cache_file, 'rt')
        else:
            input_fp = open(self.cache_file, 'r')
        self.load_str(input_fp.read())
        input_fp.close()

    def load_str(self, input_str):
        """Load subprocess output from a string

        Args:
            input_str (str): The text that came from the subprocess's stdout and
                should be parsed by this method.
        """
        self['_raw'] = input_str

    def save_cache(self, output_file=None):
        """Serialize subprocess output to a text file

        Args:
            output_file (str): Path to a file to which the output cache should
                be written.  If None, write to stdout.
        """
        if output_file is None:
            sys.stdout.write(str(self))
        else:
            with open(output_file, 'w') as output_fp:
                output_fp.write(str(self))

def walk_file_collection(input_source):
    """Walk all member files of an input source.

    Iterator that visits every member of an input source (either directory or
    tarfile) and yields its file name, last modify time, and a file handle to
    its contents.

    Args:
        input_source (str): A path to either a directory containing files or a
            tarfile containing files.

    Yields:
        tuple: Attributes for a member of `input_source` with the following
        data:

        * str: fully qualified path corresponding to its name
        * float: last modification time expressed as seconds since epoch
        * file: handle to access the member's contents
    """

    if os.path.isdir(input_source):
        for root, _, files in os.walk(input_source):
            for file_name in files:
                fq_file_name = os.path.join(root, file_name)
                yield (fq_file_name,
                       os.path.getmtime(fq_file_name),
                       open(fq_file_name, 'r'))
    else:
        _, encoding = mimetypes.guess_type(input_source)
        if encoding == 'gzip':
            file_obj = tarfile.open(input_source, 'r:gz')
        else:
            file_obj = tarfile.open(input_source, 'r')
        for member in file_obj.getmembers():
            if member.isfile():
                yield (member.name,
                       member.mtime,
                       file_obj.extractfile(member))



#!/usr/bin/env python

import sys

DEBUG = False

def debug_print(string):
    """
    Print debug messages if the module's global debug flag is enabled.
    """
    if DEBUG:
        sys.stderr.write(string + "\n")
    return

def error(string):
    """
    Handle errors generated within TOKIO.  Currently just a passthrough to
    stderr; should probably provide exceptions later on.
    """
    sys.stderr.write(string + "\n")

def warning(string):
    """
    Handle warnings generated within TOKIO.  Currently just a passthrough to
    stderr; should probably provide a more rigorous logging/reporting
    interface later on.
    """
    sys.stderr.write(string + "\n")



"""Site-independent interface to retrieve job info
"""

import os
import datetime
import tokio.config as config

try:
    import tokio.connectors.slurm
except ImportError:
    pass

try:
    import tokio.connectors.nersc_jobsdb
except ImportError:
    pass

DEFAULT_JOBID_PROVIDERS = ['slurm']
DEFAULT_JOBNODES_PROVIDERS = ['slurm']

def get_job_startend(jobid, cache_file=None):
    """Find earliest start and latest end time for a job.

    Returns:
        tuple of datetime.datetime: Two-item tuple of (earliest start time,
            latest end time)

    Raises:
        tokio.ConfigError: When no valid providers are found
    """
    jobid_providers = config.CONFIG.get('jobinfo_jobid_providers', DEFAULT_JOBID_PROVIDERS)
    for jobid_provider in jobid_providers:
        if jobid_provider == 'slurm':
            slurm_job = tokio.connectors.slurm.Slurm(jobid=jobid, cache_file=cache_file)
            return slurm_job.get_job_startend()
        elif jobid_provider == 'nersc_jobsdb':
            nersc_host = config.CONFIG.get('nersc_host')
            if nersc_host is None:
                nersc_host = os.environ.get('NERSC_HOST')
            if nersc_host is None:
                raise KeyError("NERSC_HOST not defined in environment or pytokio config")
            nersc_jobsdb = tokio.connectors.nersc_jobsdb.NerscJobsDb(cache_file=cache_file)
            start, end = nersc_jobsdb.get_job_startend(jobid=jobid, nersc_host=nersc_host)
            return datetime.datetime.fromtimestamp(start), datetime.datetime.fromtimestamp(end)
        else:
            raise tokio.ConfigError("No valid jobid providers found")
    raise tokio.ConfigError("No valid jobid providers found")

def get_job_nodes(jobid, cache_file=None):
    """Return a list of all job nodes used.

    Creates a list of all nodes used for a jobid.

    Returns:
        set: Set of node names used by the job described by this object

    Raises:
        tokio.ConfigError: When no valid providers are found
    """
    jobnodes_providers = config.CONFIG.get('jobinfo_jobnodes_providers', DEFAULT_JOBNODES_PROVIDERS)
    for jobnodes_provider in jobnodes_providers:
        if jobnodes_provider == 'slurm':
            slurm_job = tokio.connectors.slurm.Slurm(jobid=jobid, cache_file=cache_file)
            return slurm_job.get_job_nodes()
        else:
            raise tokio.ConfigError("No valid jobnodes providers found")
    raise tokio.ConfigError("No valid jobnodes providers found")



#!/usr/bin/env python
"""
Interface with an LMT database.  Provides wrappers for common queries using the
CachingDb class.
"""

import os
import datetime
from . import cachingdb

### Names and schemata of all LMT database tables worth caching
LMTDB_TABLES = {
    "FILESYSTEM_INFO": {
        'columns': [
            'FILESYSTEM_ID',
            'FILESYSTEM_NAME',
            'FILESYSTEM_MOUNT_NAME,'
            'SCHEMA_VERSION',
        ],
        'primary_key': ['FILESYSTEM_ID'],
    },
    "MDS_DATA": {
        'columns': [
            'MDS_ID',
            'TS_ID',
            'PCT_CPU',
            'KBYTES_FREE',
            'KBYTES_USED',
            'INODES_FREE',
            'INODES_USED',
        ],
        'primary_key': ['MDS_ID', 'TS_ID'],
    },
    "MDS_INFO": {
        'columns': [
            'MDS_ID',
            'FILESYSTEM_ID',
            'MDS_NAME',
            'HOSTNAME',
            'DEVICE_NAME'
        ],
        'primary_key': ['MDS_ID'],
    },
    "MDS_OPS_DATA": {
        'columns': [
            'MDS_ID',
            'TS_ID',
            'OPERATION_ID',
            'SAMPLES',
            'SUM',
            'SUMSQUARES',
        ],
        'primary_key': ['MDS_ID', 'TS_ID', 'OPERATION_ID'],
    },
    "MDS_VARIABLE_INFO": {
        'columns': [
            'VARIABLE_ID',
            'VARIABLE_NAME',
            'VARIABLE_LABEL',
            'THRESH_TYPE',
            'THRESH_VAL1',
            'THRESH_VAL2'
        ],
        'primary_key': ['VARIABLE_ID'],
    },
    "OPERATION_INFO": {
        'columns': [
            'OPERATION_ID',
            'OPERATION_NAME',
            'UNITS'
        ],
        'primary_key': ['OPERATION_ID'],
    },
    "OSS_DATA": {
        'columns': [
            'OSS_ID',
            'TS_ID',
            'PCT_CPU',
            'PCT_MEMORY'
        ],
        'primary_key': ['OSS_ID', 'TS_ID'],
    },
    "OSS_INFO": {
        'columns': [
            'OSS_ID',
            'FILESYSTEM_ID',
            'HOSTNAME',
            'FAILOVERHOST'
        ],
        'primary_key': [
            'OSS_ID',
            'HOSTNAME'
        ],
    },
    "OST_DATA": {
        'columns': [
            'OST_ID',
            'TS_ID',
            'READ_BYTES',
            'WRITE_BYTES',
            'PCT_CPU',
            'KBYTES_FREE',
            'KBYTES_USED',
            'INODES_FREE',
            'INODES_USED'
        ],
        'primary_key': ['OST_ID', 'TS_ID'],
    },
    "OST_INFO": {
        'columns': [
            'OST_ID',
            'OSS_ID',
            'OST_NAME',
            'HOSTNAME',
            'OFFLINE',
            'DEVICE_NAME'
        ],
        'primary_key': ['OST_ID'],
    },
    "OST_VARIABLE_INFO": {
        'columns': [
            'VARIABLE_ID',
            'VARIABLE_NAME',
            'VARIABLE_LABEL',
            'THRESH_TYPE',
            'THRESH_VAL1',
            'THRESH_VAL2'
        ],
        'primary_key': ['VARIABLE_ID'],
    },
    "TIMESTAMP_INFO": {
        'columns': ['TS_ID', 'TIMESTAMP'],
        'primary_key': ['TS_ID'],
    },
}

class LmtDb(cachingdb.CachingDb):
    """
    Class to wrap the connection to an LMT MySQL database or SQLite database
    """
    def __init__(self, dbhost=None, dbuser=None, dbpassword=None, dbname=None, cache_file=None):
        """
        Initialize LmtDb with either a MySQL or SQLite backend
        """
        # Get database parameters
        if dbhost is None:
            dbhost = os.environ.get('PYTOKIO_LMT_HOST')
        if dbuser is None:
            dbuser = os.environ.get('PYTOKIO_LMT_USER')
        if dbpassword is None:
            dbpassword = os.environ.get('PYTOKIO_LMT_PASSWORD')
        if dbname is None:
            dbname = os.environ.get('PYTOKIO_LMT_DB')

        super(LmtDb, self).__init__(
            dbhost=dbhost,
            dbuser=dbuser,
            dbpassword=dbpassword,
            dbname=dbname,
            cache_file=cache_file)

        # The list of OST names is an immutable property of a database, so
        # fetch and cache it here.  Also maintain a mapping of OST_ID to
        # OST_NAME so we don't have to do JOINs in SQL; this will not scale
        # to gargantuan Lustre clusters though.
        self.ost_names = []
        self.ost_id_map = {}
        for row in self.query('SELECT OST_ID, OST_NAME FROM OST_INFO'):
            self.ost_names.append(row[1])
            self.ost_id_map[row[0]] = row[1]
        self.ost_names = tuple(self.ost_names)

        # Do the same for OSSes
        self.oss_names = []
        self.oss_id_map = {}
        for row in self.query('SELECT OSS_ID, HOSTNAME FROM OSS_INFO'):
            self.oss_names.append(row[1])
            self.oss_id_map[row[0]] = row[1]
        self.oss_names = tuple(self.oss_names)

        # Do the same for MDSes
        self.mds_names = []
        self.mds_id_map = {}
        for row in self.query('SELECT MDS_ID, HOSTNAME FROM MDS_INFO'):
            self.mds_names.append(row[1])
            self.mds_id_map[row[0]] = row[1]
        self.mds_names = tuple(self.mds_names)

        # Do the same for MDS operations
        self.mds_op_names = []
        self.mds_op_id_map = {}
        for row in self.query('SELECT OPERATION_ID, OPERATION_NAME FROM OPERATION_INFO'):
            self.mds_op_names.append(row[1])
            self.mds_op_id_map[row[0]] = row[1]
        self.mds_op_names = tuple(self.mds_op_names)

    def get_ts_ids(self, datetime_start, datetime_end):
        """
        Given a starting and ending time, return the lowest and highest ts_id
        values that encompass those dates, inclusive.
        """
        # First figure out the timestamp range
        query_str = "SELECT TIMESTAMP_INFO.TS_ID FROM TIMESTAMP_INFO WHERE TIMESTAMP >= %(ps)s AND TIMESTAMP <= %(ps)s"
        query_variables = (
            datetime_start.strftime("%Y-%m-%d %H:%M:%S"),
            datetime_end.strftime("%Y-%m-%d %H:%M:%S"))

        result = self.query(query_str=query_str,
                            query_variables=query_variables)
        ts_ids = [x[0] for x in result]
        ### TODO: make sure this works (it's templated down in test_bin_cache_lmtdb.py)
        return min(ts_ids), max(ts_ids)

    def get_timeseries_data(self, table, datetime_start, datetime_end, timechunk=datetime.timedelta(hours=1)):
        """
        Break a timeseries query into smaller queries over smaller time ranges.
        This is an optimization to avoid the O(N*M) scaling of the JOINs in the
        underlying SQL query.
        """
        table_schema = LMTDB_TABLES.get(table.upper())
        if table_schema is None:
            raise KeyError("Table '%s' is not valid" % table)
        else:
            result_columns = ['TIMESTAMP'] + table_schema['columns']
        format_dict = {
            'schema': ', '.join(result_columns).replace("TS_ID,", "TIMESTAMP_INFO.TS_ID,"),
            'table': table,
        }

        index0 = len(self.saved_results.get(table, {'rows': []})['rows'])
        chunk_start = datetime_start
        while chunk_start < datetime_end:
            if timechunk is None:
                chunk_end = datetime_end
            else:
                chunk_end = chunk_start + timechunk
            if chunk_end > datetime_end:
                chunk_end = datetime_end
            start_stamp = chunk_start.strftime("%Y-%m-%d %H:%M:%S")
            end_stamp = chunk_end.strftime("%Y-%m-%d %H:%M:%S")

            query_str = """SELECT
                               %(schema)s
                           FROM
                               %(table)s
                           INNER JOIN TIMESTAMP_INFO ON TIMESTAMP_INFO.TS_ID = %(table)s.TS_ID
                           WHERE
                               TIMESTAMP_INFO.TIMESTAMP >= %%(ps)s
                               AND TIMESTAMP_INFO.TIMESTAMP < %%(ps)s
                           """ % format_dict
            self.query(query_str, (start_stamp, end_stamp), table=table, table_schema=table_schema)
            if timechunk is not None:
                chunk_start += timechunk

        return self.saved_results[table]['rows'][index0:], result_columns

    def get_mds_data(self, datetime_start, datetime_end, timechunk=datetime.timedelta(hours=1)):
        """Schema-agnostic method for retrieving MDS load data.

        Wraps get_timeseries_data() but fills in the exact table name used in
        the LMT database schema.

        Args:
            datetime_start (datetime.datetime): lower bound on time series data
                to retrieve, inclusive
            datetime_End (datetime.datetime): upper bound on time series data to
                retrieve, exclusive
            timechunk (datetime.timedelta): divide time range query into
                sub-ranges of this width to work around N*N scaling of JOINs

        Returns:
            Tuple of (results, column names)  where results are tuples of tuples
            as returned by the MySQL query and column names are the names of
            each column expressed in the individual rows of results.
        """
        return self.get_timeseries_data('MDS_DATA',
                                        datetime_start,
                                        datetime_end,
                                        timechunk=timechunk)

    def get_mds_ops_data(self, datetime_start, datetime_end, timechunk=datetime.timedelta(hours=1)):
        """Schema-agnostic method for retrieving metadata operations data.

        Wraps get_timeseries_data() but fills in the exact table name used in
        the LMT database schema.

        Args:
            datetime_start (datetime.datetime): lower bound on time series data
                to retrieve, inclusive
            datetime_End (datetime.datetime): upper bound on time series data to
                retrieve, exclusive
            timechunk (datetime.timedelta): divide time range query into
                sub-ranges of this width to work around N*N scaling of JOINs

        Returns:
            Tuple of (results, column names)  where results are tuples of tuples
            as returned by the MySQL query and column names are the names of
            each column expressed in the individual rows of results.
        """
        return self.get_timeseries_data('MDS_OPS_DATA',
                                        datetime_start,
                                        datetime_end,
                                        timechunk=timechunk)

    def get_oss_data(self, datetime_start, datetime_end, timechunk=datetime.timedelta(hours=1)):
        """Schema-agnostic method for retrieving OSS data.

        Wraps get_timeseries_data() but fills in the exact table name used in
        the LMT database schema.

        Args:
            datetime_start (datetime.datetime): lower bound on time series data
                to retrieve, inclusive
            datetime_End (datetime.datetime): upper bound on time series data to
                retrieve, exclusive
            timechunk (datetime.timedelta): divide time range query into
                sub-ranges of this width to work around N*N scaling of JOINs

        Returns:
            Tuple of (results, column names)  where results are tuples of tuples
            as returned by the MySQL query and column names are the names of
            each column expressed in the individual rows of results.
        """
        return self.get_timeseries_data('OSS_DATA',
                                        datetime_start,
                                        datetime_end,
                                        timechunk=timechunk)

    def get_ost_data(self, datetime_start, datetime_end, timechunk=datetime.timedelta(hours=1)):
        """Schema-agnostic method for retrieving OST data.

        Wraps get_timeseries_data() but fills in the exact table name used in
        the LMT database schema.

        Args:
            datetime_start (datetime.datetime): lower bound on time series data
                to retrieve, inclusive
            datetime_End (datetime.datetime): upper bound on time series data to
                retrieve, exclusive
            timechunk (datetime.timedelta): divide time range query into
                sub-ranges of this width to work around N*N scaling of JOINs

        Returns:
            Tuple of (results, column names)  where results are tuples of tuples
            as returned by the MySQL query and column names are the names of
            each column expressed in the individual rows of results.
        """
        return self.get_timeseries_data('OST_DATA',
                                        datetime_start,
                                        datetime_end,
                                        timechunk=timechunk)



"""
Given one or more Darshan logs containing both POSIX and Lustre counters,
attempt to determine the performance each file saw and try to correlate poorly
performing files with specific Lustre OSTs.

This tool first estimates per-file I/O bandwidths by dividing the total bytes
read/written to each file by the time the application spent performing I/O to
that file.  It then uses data from Darshan's Lustre module to map these
performance estimates to the OSTs over which each file was striped.
With the list of OSTs and performance measurements corresponding to each OST,
the Pearson correlation coefficient is then calculated between performance and
each individual OST.

Multiple Darshan logs can be passed to increase the number of observations used
for correlation.  This tool does not work unless the Darshan log(s) contain
data from the Lustre module.
"""

import json
import argparse
import warnings
import pandas
import scipy.stats
import tokio.connectors.darshan

# OST_NAME_FMT = "OST%05x"
OST_NAME_FMT = "OST#%d"

def correlate_ost_performance(darshan_logs):
    """
    Generate a DataFrame containing files, performance measurements, and OST
    mappings and attempt to correlate performance with individual OSTs.
    """
    darshan_df = darshanlogs_to_ost_dataframe(darshan_logs)

    # in the unlikely event that all performance measurements are the same (or
    # there is only one), we simply cannot perform correlation analysis
    if len(darshan_df['performance'].unique()) == 1:
        return []

    results = []
    for ost_name in [x for x in darshan_df.columns if x.startswith('OST')]:
        # if all values are the same (or there is only one), we cannot correlate
        if len(darshan_df[ost_name].unique()) == 1:
            continue
        coeff, pval = scipy.stats.pearsonr(darshan_df['performance'],
                                           darshan_df[ost_name])
        results.append({
            'ost_name': ost_name,
            'coefficient': coeff,
            'p-value': pval})
    return results

def estimate_darshan_perf(ranks_data):
    """
    Calculate performance in a sideways fashion: find the longest I/O time
    across any rank for this file, then divide the sum of all bytes
    read/written by this longest io time.  This function expects to receive
    a dict that is keyed by MPI ranks (or a single "-1" key) and whose
    values are dicts corresponding to Darshan POSIX counters.
    """
    max_io_time = 0.0
    sum_bytes = 0.0
    for counter_data in ranks_data.values():
        this_io_time = counter_data['F_WRITE_TIME'] + \
                       counter_data['F_READ_TIME'] + \
                       counter_data['F_META_TIME']
        if this_io_time > max_io_time:
            max_io_time = this_io_time
        sum_bytes += counter_data['BYTES_READ'] + \
                     counter_data['BYTES_WRITTEN']
    return sum_bytes / max_io_time

def summarize_darshan_perf(darshan_logs):
    """
    Given a list of Darshan log file paths, calculate the performance observed
    from each file and identify OSTs over which each file was striped.  Return
    this summary of file performances and stripes.
    """
    results = {
        'file_paths': [],
        'performance': [],
        'ost_lists': [],
    }
    for darshan_log in darshan_logs:
        darshan_data = tokio.connectors.darshan.Darshan(darshan_log)
        darshan_data.darshan_parser_base()
        if 'counters' not in darshan_data:
            warnings.warn("Invalid Darshan log %s" % darshan_log)
            continue
        elif 'posix' not in darshan_data['counters']:
            warnings.warn("Darshan log %s does not contain POSIX module data" % darshan_log)
            continue
        elif 'lustre' not in darshan_data['counters']:
            warnings.warn("Darshan log %s does not contain Lustre module data" % darshan_log)
            continue
        counters = darshan_data['counters']

        for logged_file_path, ranks_data in counters['posix'].items():
            # encode the darshan log's name in addition to the file path in case
            # multiple Darshan logs with identical file paths (but different
            # striping) are being processed
            file_path = "%s@%s" % (darshan_log, logged_file_path)

            # calculate the file's I/O performance
            performance = estimate_darshan_perf(ranks_data)

            # assemble a list of OSTs
            ost_list = set([])
            if logged_file_path not in counters['lustre']:
                continue
            for counter_data in counters['lustre'][logged_file_path].values():
                for ost_id in range(counter_data['STRIPE_WIDTH']):
                    key = "OST_ID_%d" % ost_id
                    ost_list.add(counter_data[key])

            # append findings from this file record to the master dict
            if file_path not in results['file_paths']:
                results['file_paths'].append(file_path)
                results['performance'].append(performance)
                results['ost_lists'].append(list(ost_list))
            else:
                index = results['file_paths'].index(file_path)
                if results['performance'][index] < performance:
                    results['performance'][index] = performance
                results['ost_lists'][index] = \
                    list(set(results['ost_lists'][index]) | ost_list)
    return results

def darshanlogs_to_ost_dataframe(darshan_logs):
    """
    Given a set of Darshan log file paths, create a dataframe containing each
    file, its observed performance, and a matrix of values corresponding to what
    fraction of that file's contents were probably striped on each OST.
    """

    # get a dict of files, their performances, and their stripes
    results = summarize_darshan_perf(darshan_logs)

    # build a dataframe from the results dictionary
    pre_dataframe = {
        'file_paths': results['file_paths'],
        'performance': results['performance'],
    }

    # create one column for each OST
    for ost_id_list in results['ost_lists']:
        for ost_id in ost_id_list:
            ost_name = OST_NAME_FMT % ost_id
            if ost_name not in pre_dataframe:
                pre_dataframe[ost_name] = [0.0] * len(results['file_paths'])

    # for each file record, calculate the fraction it contributed to each OST
    for index in range(len(results['file_paths'])):
        num_osts = float(len(results['ost_lists'][index]))
        for ost_id in results['ost_lists'][index]:
            ost_name = OST_NAME_FMT % ost_id
            pre_dataframe[ost_name][index] = 1.0 / num_osts

    return pandas.DataFrame(pre_dataframe)

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-j", "--json", action="store_true", help="return output in JSON format")
    parser.add_argument("-p", "--p-threshold", type=float, default=0.05,
                        help="p-value above which correlations will not be displayed")
    parser.add_argument("-c", "--c-threshold", type=float, default=0.0,
                        help="coefficient below which abs(correlations) will not be displayed")
    parser.add_argument("-C", "--c-warning", type=float, default=0.5,
                        help="print a warning if abs(correlations) is above this threshold")
    parser.add_argument("darshanlogs", nargs="*", default=None, help="darshan logs to process")
    args = parser.parse_args(argv)

    results = correlate_ost_performance(args.darshanlogs)

    if args.json:
        filtered_results = []
        for result in sorted(results, key=lambda k: k['coefficient']):
            if result['p-value'] < args.p_threshold \
            and abs(result['coefficient']) > args.c_threshold:
                filtered_results.append(result)
        print(json.dumps(filtered_results, indent=4, sort_keys=True))
    else:
        print("%-10s %12s %10s" % ("OST Name", "Correlation", "P-Value"))
        for result in sorted(results, key=lambda k: k['coefficient']):
            if result['p-value'] < args.p_threshold \
            and abs(result['coefficient']) > args.c_threshold:
                if abs(result['coefficient']) > args.c_warning:
                    if result['coefficient'] < 0:
                        result['warning'] = " <-- this OST looks unhealthy"
                    else:
                        result['warning'] = " <-- this OST is running unusually quickly"
                else:
                    result['warning'] = ""
                print("%(ost_name)-10s %(coefficient)12.3f %(p-value)10.4g%(warning)s" % result)
